
from kivy.properties import ListProperty, NumericProperty
from kivy.lang import Builder
from kivy.animation import Animation
from kivy.clock import Clock
from kivy.uix.label import Label

from kivygo.behaviors.hover import HoverBehavior
from kivygo.behaviors.touch_effecs import EffectBehavior
from kivygo.behaviors.button import ButtonBehavior, ToggleButtonBehavior


Builder.load_string("""

<ButtonEffect>:
	canvas.before:
		Color:
			rgba: self.background
		RoundedRectangle:
			size: self.size
			pos: self.pos
			radius: self.radius
	canvas.after:
		Color:
			rgba: self.background_line
		Line:
			rounded_rectangle: (self.pos + self.size + self.radius + [100])
			width: self.width_line

			
<ToggleButtonEffect>:
	

""")

class ButtonEffect(ButtonBehavior, Label, EffectBehavior, HoverBehavior):
	#Colors
	background_line = ListProperty([0, 0, 0, 0])
	background = ListProperty([0, 0, 0, 0])
	
	down_color_text = ListProperty([0, 0, 0, 0])
	down_color_line = ListProperty([0, 0, 0, 0])

	radius = ListProperty([0, 0, 0, 0])

	down_background_color = ListProperty([])
	background_color = ListProperty([[0.05, 0.0, 0.4, 1], [0.0625, 0.0, 0.5, 1]])
	color_line = ListProperty([[1, 1, 1, 1], [0.8, 0.925, 1, 1]])
	color_text = ListProperty([[1, 1, 1, 1], [0.8, 0.925, 1, 1]])

	duration_back = NumericProperty(0.2)

	#Sizes
	width_line = NumericProperty(1.01)
	_pressed = False

	def __init__(self, **kwargs):
		self.bind(
			down_background_color=self.set_color,
			background_color=self.set_color,
			color_line=self.set_color,
			color_text=self.set_color)
		self.type_button = 'rounded'
		super().__init__(**kwargs)
		Clock.schedule_once(self.set_color)
	
	def set_color(self, *args):
		if self.down_background_color:
			self.background = self.get_color(self.down_background_color, 0)
		else:
			self.background = self.get_color(self.background_color, 0)
		
		self.background_line = self.get_color(self.color_line, 0)
		self.color = self.get_color(self.color_text, 0)
	
	def get_color(self, object, index):
		if isinstance(object, (list, tuple)):
			if len(object) == 2 and index > -1 and index < 3:
				return object[index]
		return object


	def set_pressed(self, *args):
		if self.down_color_text[-1] != 0:
			self.color = self.down_color_text
		
		if self.down_color_line[-1] != 0:
			self.background_line = self.down_color_line
		
		if self.down_background_color:
			self.background = self.get_color(self.down_background_color, 1)

	def on_touch_down(self, touch):
		
		if not self.collide_point(*touch.pos):
			return False

		Animation.cancel_all(self, 'color', 'background_line')
		
		touch.grab(self)
		self.ripple_show(touch)
		self._pressed = True
		result = super().on_touch_down(touch)
		if self.state == "down":
			self.set_pressed()
		return result

	def on_touch_up(self, touch):

		if touch.grab_current is self:
			touch.ungrab(self)
			self.ripple_fade()
		
		result = super().on_touch_up(touch)

		if self.down_background_color and self.state == "normal":
			self.background = self.get_color(self.down_background_color, 0)
		
		if not self.collide_point(*touch.pos):
			return False
		
		anim = Animation(
			color=self.get_color(self.color_text, 1),
			background_line=self.get_color(self.color_line, 1),
			t=self.transition_out,
			duration=self.duration_out)
		anim.bind(on_complete=self.on_touch_anim_end)
		anim.start(self)

		return result

	def on_touch_anim_end(self, *args):
		self._pressed = False
		self.new_dispatch()
		self.hover_visible = True

	def on_cursor_enter(self, *args):
		print("ON_ENTER")
		if not self._pressed:
			self.background_line = self.get_color(self.color_line, 1)
			self.color = self.get_color(self.color_text, 1)
		
		if not self.down_background_color:
			anim = Animation(
				background=self.get_color(self.background_color, 1),
				d=self.duration_back, t='out_quad')
			anim.start(self)
		return super().on_cursor_enter(*args)

	def on_cursor_leave(self, *args):
		print("ON_LEAVE")
		if not self.down_background_color:
			anim = Animation(
				background=self.get_color(self.background_color, 0),
				d=self.duration_back, t='out_quad')
			anim.start(self)
		
		if not self._pressed:
			self.color = self.get_color(self.color_text, 0)
			self.background_line = self.get_color(self.color_line, 0)
		return super().on_cursor_leave(*args)


class ToggleButtonEffect(ToggleButtonBehavior, ButtonEffect):
	pass



from kivy.uix.screenmanager import Screen
from kivy.properties import NumericProperty


class SwapScreen(Screen):

	touch_x = NumericProperty(0)
		
	def on_touch_up(self, touch):
		'''
		Chamada automaticamente pelo kivy,
		serve para mudar de tela de acordo com a posição do toque
		e o movimento.

		Args:
			touch: objeto do kivy que tem as posições do toque

		Returns:
			bool: True ou super().on_touch_up(touch) se o touch foi no widget False se não
		'''
		if not self.collide_point(*touch.pos) or self.touch_x == 0:
			return False
		
		manager = self.parent
		screen_names = manager.screen_names

		if (self.touch_x - touch.x) >= (self.width / 2.5):

			manager.transition.direction = 'left'

			if manager.next() == 'opicional':
				manager.current = screen_names[0]
			else:
				manager.current = manager.next()

		elif (self.touch_x - touch.x) <= -(self.width / 2.5):

			manager.transition.direction = 'right'

			if manager.previous() == 'opicional':
				manager.current = screen_names[len(screen_names)-2]
			else:
				manager.current = manager.previous()
		
		return super().on_touch_up(touch)

	def on_touch_down(self, touch):
		# define o valor de touch_x para ser usado em **on_touch_up**
		if self.collide_point(*touch.pos):
			self.touch_x = touch.x
			return super().on_touch_down(touch)
		
		self.touch_x = 0
		return False





import math
from kivy.properties import StringProperty
from kivy.properties import ListProperty
from kivy.properties import NumericProperty
from kivy.uix.relativelayout import RelativeLayout
from kivy.graphics import Color
from kivy.graphics import Ellipse
from kivy.graphics import Line
from kivy.uix.label import Label
from kivy.utils import get_color_from_hex


class Pizza(RelativeLayout):
    serie = ListProperty([])
    chart_size = NumericProperty(256)
    legend_color = StringProperty('ffffcc')
    legend_value_rayon = NumericProperty(100)
    legend_title_rayon = NumericProperty(160)
    chart_border = NumericProperty(2)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.chart_center = self.chart_size / 2.0

        self.bind(
            pos=self.update_pizza,
            size=self.update_pizza,
            serie=self.update_pizza,
            chart_size=self.update_pizza,
            legend_color=self.update_pizza,
            legend_value_rayon=self.update_pizza,
            legend_title_rayon=self.update_pizza,
            chart_border=self.update_pizza
        )
        
        self.bind(
            pos=self.update_label,
            size=self.update_label,
            serie=self.update_label,
            chart_size=self.update_label,
            legend_color=self.update_label,
            legend_value_rayon=self.update_label,
            legend_title_rayon=self.update_label,
            chart_border=self.update_label
        )

    def update_pizza(self, *args):
        with self.canvas:
            self.canvas.clear()
            offset_rotation = 0  # In degrees

            # Fix legend color
            Color(
                get_color_from_hex(self.legend_color)[0],
                get_color_from_hex(self.legend_color)[1],
                get_color_from_hex(self.legend_color)[2], 100
            )

            # Draw pie chart border circle
            border_circle = Line(circle=(
                self.chart_center,
                self.chart_center,
                self.chart_center),
                width=self.chart_border)

            for title, value, color in self.serie:
                angle = math.radians(((value * 3.6) / 2.0) + offset_rotation)
                title_x_pt = (math.sin(angle)) * self.legend_title_rayon
                title_y_pt = (math.cos(angle)) * self.legend_title_rayon

                # Fix color for each zone
                Color(
                    get_color_from_hex(color)[0],
                    get_color_from_hex(color)[1],
                    get_color_from_hex(color)[2], 100
                )

                # Draw zone animation
                zone = Ellipse(
                    size=(self.chart_size, self.chart_size),
                    segments=value * 3.6,
                    angle_start=offset_rotation,
                    angle_end=offset_rotation + (value * 3.6), t='in_quad'
                )

                # Offset control of each zone, drawing starts on offset
                offset_rotation += value * 3.6

    def update_label(self, *args):
        self.clear_widgets()  # Clean widget tree
        offset_rotation = 0  # In degrees

        for title, value, color in self.serie:
            angle = math.radians(((value * 3.6) / 2.0) + offset_rotation)
            value_x_pt = (math.sin(angle)) * self.legend_value_rayon
            value_y_pt = (math.cos(angle)) * self.legend_value_rayon
            title_x_pt = (math.sin(angle)) * self.legend_title_rayon
            title_y_pt = (math.cos(angle)) * self.legend_title_rayon

            # Title
            self.add_widget(Label(
                size_hint=(None, None),
                text="[color=" + self.legend_color + "]" +
                title + "[/color]",
                center_x=self.chart_center + title_x_pt,
                center_y=self.chart_center + title_y_pt,
                markup=True)
            )

            # Value
            self.add_widget(Label(
                size_hint=(None, None),
                text="[color=" + self.legend_color + "]" +
                str(value) + "[/color]",
                center_x=self.chart_center + value_x_pt,
                center_y=self.chart_center + value_y_pt,
                markup=True)
            )

            offset_rotation += value * 3.6



from kivy.properties import StringProperty, ColorProperty, OptionProperty
from kivy.uix.screenmanager import ShaderTransition
from os import path

PATH = path.dirname(__file__)
EXTRA = path.join(PATH, "extra")


class CustomTransition(ShaderTransition):
    direction = OptionProperty("lr", options=["lr", "rl"])

    glsl_file = StringProperty("Bounce.glsl")

    clearcolor = ColorProperty([0, 0, 0, 0])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        with open(path.join(EXTRA, "header.glsl")) as header, open(
            path.join(PATH, self.glsl_file)
        ) as glsl, open(path.join(EXTRA, "footer.glsl")) as footer:
            self.fs = header.read() + glsl.read() + footer.read()

    def add_screen(self, screen):
        super().add_screen(screen)
        self.render_ctx["resolution"] = list(map(float, screen.size))

        aspect_ratio = screen.size[0] / screen.size[1]

        if aspect_ratio >= 1:
            self.render_ctx["aspect"] = 1.0

        else:
            self.render_ctx["aspect"] = 2.0

        if self.direction == "lr":
            self.render_ctx["direction"] = 1.0

        else:
            self.render_ctx["direction"] = 2.0




from time import perf_counter as now

from kivy.clock import Clock
from kivy.core.window import Window
from kivy.graphics import (
    BindTexture, ClearBuffers,
    ClearColor, Color, Fbo,
    Rectangle, RenderContext,
    RoundedRectangle, Scale,
    SmoothLine, Translate,
)
from kivy.metrics import dp
from kivy.properties import (
    ColorProperty, ListProperty,
    NumericProperty, ObjectProperty,
)
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.image import Image
from kivy.uix.modalview import ModalView
from kivy.uix.screenmanager import Screen
from kivy.uix.scrollview import ScrollView
from kivy.uix.video import Video

MEAN_RES = (Window.width + Window.height) / 2


vertical_blur_shader = """

#ifdef GL_ES
    precision lowp float;
#endif

/* Outputs from the vertex shader */
varying vec4 frag_color;
varying vec2 tex_coord0;

/* uniform texture samplers */
uniform sampler2D texture0;

uniform float mean_res;
uniform float blur_size;

vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
{{
    float dt = ((blur_size / 2.0) * 1.0 / mean_res);
    vec4 sum = vec4(0.0);

    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y+3.0*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y+2.5*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y+2.0*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y+1.5*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y+1.0*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y+0.5*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y-0.5*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y-1.0*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y-1.5*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y-2.0*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y-2.5*dt))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y-3.0*dt))*0.077;

    return frag_color * vec4(sum.rgba);
}}

void main (void){
  vec4 normal_color = frag_color * texture2D(texture0, tex_coord0);
  vec4 effect_color = effect(
      normal_color, texture0, tex_coord0, gl_FragCoord.xy
  );
  gl_FragColor = effect_color;
}
"""


horizontal_blur_shader = """

#ifdef GL_ES
    precision lowp float;
#endif

/* Outputs from the vertex shader */
varying vec4 frag_color;
varying vec2 tex_coord0;

/* uniform texture samplers */
uniform sampler2D texture0;

uniform float mean_res;
uniform float blur_size;

vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
{{
    float dt = (blur_size / 2.0) * 1.0 / mean_res;
    vec4 sum = vec4(0.0);

    sum += texture2D(texture, vec2(tex_coords.x+3.0*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x+2.5*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x+2.0*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x+1.5*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x+1.0*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x+0.5*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x-0.5*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x-1.0*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x-1.5*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x-2.0*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x-2.5*dt, tex_coords.y))*0.077;
    sum += texture2D(texture, vec2(tex_coords.x-3.0*dt, tex_coords.y))*0.077;

    return  frag_color * vec4(sum.rgba);
}}

void main (void){
  vec4 normal_color = frag_color * texture2D(texture0, tex_coord0);
  vec4 effect_color = effect(
      normal_color, texture0, tex_coord0, gl_FragCoord.xy
  );
  gl_FragColor = effect_color;
}
"""


noise_shader = """

#ifdef GL_ES
    precision highp float;
#endif

/* Outputs from the vertex shader */
varying vec4 frag_color;
varying vec2 tex_coord0;

/* uniform texture samplers */
uniform sampler2D texture0;


vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
{{
    vec4 sum = vec4(0.0);
    float a = 12.9898;
    float b = 78.233;
    float c = 43758.5453;
    float dt2 = dot(tex_coords.xy, vec2(a, b));
    float sn= mod(dt2, 3.14);
    vec4 noise = vec4(fract(sin(sn) * c));

    vec4 bg = texture2D(texture0, tex_coord0);
    sum = mix(bg, vec4(noise.xyz, 1.0), 1.0);

    return vec4(sum.rgba);
}}

void main (void){
  vec4 normal_color = frag_color * texture2D(texture0, tex_coord0);
  vec4 effect_color = effect(
      normal_color, texture0, tex_coord0, gl_FragCoord.xy
  );
  gl_FragColor = effect_color;
}
"""


final_shader_effect = """

#ifdef GL_ES
    precision highp float;
#endif

$HEADER$

uniform int add_child_texture;
uniform float opacity;
uniform float luminosity;
uniform float saturation;
uniform float noise_opacity;
uniform vec2 position;
uniform vec2 resolution;
uniform vec4 color_overlay;
uniform sampler2D texture1;
uniform sampler2D texture2;

void main(void)
{
    vec2 pos = (gl_FragCoord.xy - position.xy) / resolution.xy;
    float noise = 0.0;
    vec4 textureResult = vec4(0.0);
    vec4 effect_texture = texture2D(texture1, pos);
    vec4 noise_texture = texture2D(texture2, tex_coord0);

    const vec3 W = vec3(0.2125, 0.7154, 0.0721);
    vec3 intensity = vec3(dot(effect_texture.rgb, W));
    effect_texture = vec4(mix(intensity, effect_texture.rgb, saturation), 1.0);
    effect_texture *= luminosity;

    effect_texture = mix(
        effect_texture.rgba,
        vec4(color_overlay.rgb, 1.0),
        min(1.0, color_overlay.a)
    );
    effect_texture = mix(
        effect_texture.rgba,
        vec4(noise_texture.rgb, 1.0),
        min(1.0, noise_opacity)
    );

    gl_FragColor = vec4(effect_texture.rgb, opacity);
}
"""


class VerticalBlur(Fbo):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, fs=vertical_blur_shader, **kwargs)
        self["mean_res"] = MEAN_RES
        self["blur_size"] = dp(25)


class HorizontalBlur(Fbo):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, fs=horizontal_blur_shader, **kwargs)
        self.rect = Rectangle()
        self["mean_res"] = MEAN_RES
        self["blur_size"] = dp(25)


class Noise(Fbo):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, fs=noise_shader, **kwargs)
        self.rect = Rectangle()


class FrostedGlass(FloatLayout):

    background = ObjectProperty(None, allownone=True)
    """
    Target widget/layout that will be used as a background to FrostedGlass.
    The recomended way to pass the widget is through the id.
    """

    blur_size = NumericProperty(25)
    """
    Size of the gaussian blur aplied to the background.

    Note: Do not pass relative values such as dp or sp. FrostedGlass already
    manages this automatically, according to the device's screen density.
    """

    noise_opacity = NumericProperty(0.1)
    """Opacity of the noise layer.
    """

    saturation = NumericProperty(1.2)
    """Saturation boost that will be aplied to the background.
    """

    luminosity = NumericProperty(1.3)
    """Luminosity boost that will be aplied to the background.
    """

    overlay_color = ColorProperty([0.5, 0.5, 0.5, 0.5])
    """Color/tint overlay that will be aplied over the background.
    """

    border_radius = ListProperty([0, 0, 0, 0])
    """Specifies the radius used for the rounded corners clockwise:
    top-left, top-right, bottom-right, bottom-left.
    """

    outline_color = ColorProperty([1.0, 1.0, 1.0, 1.0])
    """FrostedGlass outline color.
    """

    outline_width = NumericProperty(1)
    """FrostedGlass outline width.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        fbind = self.fbind
        fbind("outline_width", self._update_canvas)
        fbind("outline_color", self._update_canvas)
        fbind("border_radius", self._update_canvas)
        fbind("noise_opacity", self.update_effect)
        fbind("luminosity", self.update_effect)
        fbind("saturation", self.update_effect)
        fbind("overlay_color", self.update_effect)
        fbind("border_radius", self.update_effect)

        self.frosted_glass_effect = RenderContext(
            use_parent_projection=True,
            use_parent_modelview=True,
            fs=final_shader_effect
        )
        with self.frosted_glass_effect:
            self.bt_1 = BindTexture(index=1)
            self.bt_2 = BindTexture(index=2)
            self.fbo_rect = RoundedRectangle(
                size=self.size,
                pos=self.pos,
                radius=self.border_radius,
            )

        self.frosted_glass_effect["texture1"] = 1
        self.frosted_glass_effect["texture2"] = 2
        self.canvas.add(self.frosted_glass_effect)

        with self.canvas:
            self._outline_color = Color(rgba=self.outline_color)
            self.outline = SmoothLine(
                width=1,
                overdraw_width=2,
                rounded_rectangle=(
                    self.x, self.y,
                    self.width, self.height,
                    1, 1, 1, 1, 45,
                ),
            )

        self.noise = Noise(size=(100, 100))
        self.h_blur = HorizontalBlur(size=(100, 100))
        self.v_blur = VerticalBlur(size=(100, 100))

        with self.h_blur:
            ClearColor(0, 0, 0, 0)
            ClearBuffers()
            self.h_blur_scale = Scale(1, 1, 1)
            self.h_blur_translate = Translate(0, 0)

        with self.v_blur:
            ClearColor(0, 0, 0, 0)
            ClearBuffers()
            self.v_blur_scale = Scale(1, 1, 1)
            self.v_blur_translate = Translate(0, 0)

        self.last_value = 0
        self.last_update_time = 0
        self.last_value_list = [0, 0]
        self.last_blur_size_value = None
        self.last_fbo_pos = [None, None]
        self._pos = [0, 0]
        self.popup_parent = None
        self.parent_screen = None
        self.parents_list = []
        self.background_children_list = []
        self.background_parents_list = []

        self.is_movable = False
        self.adapted_fbo_size = False

        self._update_glsl_ev = Clock.create_trigger(self._update_glsl, 0)
        self._update_fbo_ev = Clock.create_trigger(self._update_fbo_effect, 0)
        self._update_texture_ev = Clock.create_trigger(self._set_final_texture, 0)
        self._refresh_effect_ev = Clock.create_trigger(self.refresh_effect, 0.033333, True)

    def update_effect(self, *args):
        self._update_glsl_ev()

    def refresh_effect(self, *args):
        self._update_fbo_ev()
        self._update_glsl_ev()

    def _update_glsl(self, *args):
        self._pos = self.to_window(*self.pos)
        if self.not_current_screen \
            or self.out_of_the_window \
            and self.background_loaded:
            return None

        effect = self.frosted_glass_effect
        effect["position"] = list(map(float, self._pos))
        effect["resolution"] = list(map(float, self.size))
        effect["luminosity"] = float(self.luminosity)
        effect["saturation"] = float(self.saturation)
        effect["noise_opacity"] = float(self.noise_opacity)
        effect["color_overlay"] = list(map(float, self.overlay_color))

        if not self.is_movable:
            self._update_texture_ev()
            return None
        
        if not self.adapted_fbo_size:
            self.refresh_effect()
            self.adapted_fbo_size = True

        self.h_blur_translate.x = self.v_blur_translate.x = -self._pos[0]
        self.h_blur_translate.y = self.v_blur_translate.y = (-self._pos[1] - self.size[1])

    def _set_final_texture(self, pos):
        if not self.background:
            return None

        if self.background.canvas not in self.h_blur.children:
            self.h_blur.add(self.background.canvas)
            self.h_blur.draw()
            self.v_blur.add(self.h_blur.rect)
            self.v_blur.draw()

        if self.h_blur.rect.texture != self.h_blur.texture:
            self.h_blur.rect.texture = self.h_blur.texture

        self.h_blur.rect.size = self.size
        self.h_blur.rect.pos = self._pos

        self.h_blur.draw()
        self.v_blur.draw()
        self.v_blur.ask_update()
        Window.canvas.ask_update()

        self.bt_1.texture = self.v_blur.texture

        if self._update_texture_ev.timeout == 0:
            self._update_texture_ev.timeout = -1

    def _update_noise_texture(self):
        fbo_size = max(1, self.width / dp(1)), max(1, self.height / dp(1))
        self.noise.size = fbo_size
        self.noise.rect.size = self.size
        self.noise.add(self.noise.rect)
        self.noise.draw()
        self.noise.remove(self.noise.rect)
        self.bt_2.texture = self.noise.texture

    def _update_fbo_effect(self, *args):
        if self.is_movable:
            fbo_size = (min(self.width, 150), min(self.height, 150))
        else:
            fbo_size = (min(self.width, 250), min(self.height, 250))

        size = max(1, self.width), max(1, self.height)
        fbo_size = max(1, fbo_size[0]), max(1, fbo_size[1])

        self.h_blur.size = fbo_size
        self.v_blur.size = fbo_size

        pos = self.to_window(*self.pos)
        x = 1 / (size[0] / fbo_size[0])
        y = 1 / (size[1] / fbo_size[1])

        self.h_blur_scale.x = self.v_blur_scale.x = x
        self.h_blur_scale.y = self.v_blur_scale.y = -y
        self.h_blur_translate.x = self.v_blur_translate.x = -pos[0]
        self.h_blur_translate.y = self.v_blur_translate.y = -pos[1] - size[1]

    def on_touch_down(self, touch):
        if self.collide_point(*touch.pos):
            super().on_touch_down(touch)
            return True
        
        return super().on_touch_down(touch)

    def on_size(self, instance, size):
        self._update_canvas()
        self._update_noise_texture()
        self.refresh_effect()

    def on_pos(self, *args):
        self._update_canvas()
        self.refresh_effect()

    def _update_canvas(self, *args):
        def modify_border(round_number):
            return max(1, min(min(self.width, self.height) / 2, round_number))
        
        border_radius = list(map(modify_border, self.border_radius))
        self.fbo_rect.size = self.size
        self.fbo_rect.pos = self.pos
        self.fbo_rect.radius = border_radius

        self._outline_color.rgba = self.outline_color
        self.outline.width = self.outline_width
        self.outline.rounded_rectangle = (
            *self.pos, *self.size,
            *reversed(border_radius), 45,
        )

    def on_blur_size(self, instance, blur_size):
        blur_size = int(blur_size)
        if blur_size != self.last_blur_size_value:
            self.v_blur["blur_size"] = dp(blur_size)
            self.h_blur["blur_size"] = dp(blur_size)
            self.update_effect()
            self.last_blur_size_value = blur_size

    def on_background(self, *args):
        if not self.background:
            return None
        
        self.background_parents_list = self._get_all_parents(self.background)
        self.background_children_list = self._get_all_children(self.background)
        self._bind_parent_properties(self.background_parents_list)
        self._bind_children_properties(self.background_children_list)

    def on_parent(self, *args):
        self.parents_list = self._get_all_parents(self.parent)
        for p in self.parents_list:
            if isinstance(p, ModalView):
                self.popup_parent = p
            elif isinstance(p, Screen):
                self.parent_screen = p
            elif isinstance(p, ScrollView):
                self.is_movable = True
        
        self._bind_parent_properties(self.parents_list)

    def _get_all_parents(self, widget):
        widgets_list = []
        parent = self.parent
        
        while True:
            widgets_list.append(parent)
            if parent.parent and parent != parent.parent:
                parent = parent.parent
            else:
                return widgets_list

    def _get_all_children(self, widget):
        widgets_list = [widget]
        children_widgets = [widget]

        while children_widgets:
            parent_widgets = children_widgets
            children_widgets = []
            for w in parent_widgets:
                if w.children:
                    widgets_list.extend(w.children)
                    children_widgets.extend(w.children)
        
        return widgets_list

    def _bind_children_properties(self, children_list):
        properties_to_bind = (
            "pos",
            "size",
            "scroll_x",
            "scroll_y",
            "on_open",
            "on_enter",
            "texture",
        )
        for widget in children_list:
            for property in properties_to_bind:
                try:
                    if property == "texture":
                        if isinstance(widget, Image):
                            widget.bind(
                                texture=lambda *args:
                                    self._trigger_update_effect()
                            )
                        if isinstance(widget, Video):
                            widget.bind(
                                position=lambda *args:
                                    self._trigger_update_effect()
                            )
                    elif hasattr(widget, property):
                        widget.fbind(
                            property,
                            lambda *args: self._trigger_update_effect(),
                        )
                except Exception as e:
                    print(e)

    def _bind_parent_properties(self, parents_list):
        for widget in parents_list:

            if isinstance(widget, ScrollView):
                widget.bind(
                    size=lambda _, size: self._trigger_update_effect(size),
                    pos=lambda _, pos: self._trigger_update_effect(pos),
                    scroll_x=lambda _, scroll_x: self._trigger_update_effect(
                        scroll_x
                    ),
                    scroll_y=lambda _, scroll_y: self._trigger_update_effect(
                        scroll_y
                    ),
                )
            if isinstance(widget, ModalView):
                widget.bind(
                    on_pre_open=lambda *args: self._trigger_update_effect()
                )
            elif isinstance(widget, Screen):
                widget.bind(
                    on_pre_enter=lambda *args: self._refresh_effect_ev()
                )
                widget.bind(
                    on_enter=lambda *args: self._refresh_effect_ev.cancel()
                )

            else:
                widget.bind(
                    size=lambda _, size: self._trigger_update_effect(size)
                )
                try:
                    widget.bind(
                        pos=lambda _, pos: self._trigger_update_effect(pos)
                    )
                except Exception:
                    pass

    def _trigger_update_effect(self, value=None):
        if value == None and self.update_by_timeout:
            self.update_effect()

        if (
            (isinstance(value, int) or isinstance(value, float))
            and self.update_by_timeout
            and round(value, 3) != self.last_value
        ):
            self.update_effect()
            self.last_value = round(value, 3)

        elif (
            isinstance(value, list)
            and self.update_by_timeout
            and (
                round(value[0], 2) != self.last_value_list[0]
                or round(value[1], 2) != self.last_value_list[1]
            )
        ):
            self.update_effect()
            self.last_value_list = round(value[0], 2), round(value[1], 2)

    @property
    def popup_closed(self):
        return self.popup_parent and not self.popup_parent.parent

    @property
    def not_current_screen(self):
        if self.parent_screen == None:
            return False
        return self.parent_screen.manager.current != self.parent_screen.name

    @property
    def out_of_the_window(self):
        x, y = self.to_window(self.x, self.y)
        right, top = self.to_window(self.right, self.top)
        return right < 0 or top < 0 or x > Window.width or y > Window.height

    @property
    def update_by_timeout(self):
        _now = now()
        if _now - self.last_update_time >= 0.016666:
            self.last_update_time = _now
            return True
        return False

    @property
    def background_loaded(self):
        if not self.background:
            return False
        return self.background.canvas in self.h_blur.children



from kivy.clock import Clock
import numpy as np
from kivygo.uix.bezier import Bezier
from kivy.lang.builder import Builder
from kivy.properties import (
	NumericProperty, ListProperty,
	ObjectProperty, ColorProperty
)
from kivy.uix.floatlayout import FloatLayout
from kivy.animation import Animation
from kivy.metrics import dp


Builder.load_string("""

<CurveLayout>:

	BoxLayout:
		id: backlayout

	BoxLayout:
		canvas:
			Color:
				rgba: root.color
			Mesh:
				mode: "triangle_fan"
				vertices: root.vertices
				indices: root.indices
				texture: root.mesh_texture

	BoxLayout:
		id: frontlayout

""")


class FrontLayout(FloatLayout):
	pass


class BackLayout(FloatLayout):
	pass


class CurveLayout(FloatLayout):

	drag_distance = NumericProperty("100dp")
	curve_pos = ListProperty([0, 0])
	start_pos = ListProperty([0, 0])
	end_pos = ListProperty([0, 1000])
	color = ColorProperty([1, 1, 1, 1])
	mesh_texture = ObjectProperty(None)
	indices = ListProperty([])
	vertices = ListProperty([])

	_state = "left"
	_offset = dp(1000)


	def on_right_finish(self, *args):
		pass

	def on_right_start(self, *args):
		pass

	def on_right_second_start(self, *args):
		pass

	def on_left_finish(self, *args):
		pass

	def on_left_start(self, *args):
		pass

	def on_reset_left(self, *args):
		pass

	def on_reset_right(self, *args):
		pass

	def on_left_second_start(self, *args):
		pass


	def __init__(self, **kwargs):
		super().__init__(**kwargs)

		self.register_event_type("on_right_finish")
		self.register_event_type("on_right_start")
		self.register_event_type("on_right_second_start")
		self.register_event_type("on_left_finish")
		self.register_event_type("on_left_start")
		self.register_event_type("on_left_second_start")
		self.register_event_type("on_reset_left")
		self.register_event_type("on_reset_right")
		Clock.schedule_once(self._update)

	def _update(self, *args):
		if self._state == "left":
			self.dispatch("on_left_start")
		elif self._state == "right":
			self.dispatch("on_right_start")

	def update_mesh(self, start, end, pos):
		res = []

		points_1 = [
			[-self._offset, -self._offset],
			start,
			[pos[0] / 2, pos[1] / 2],
			pos,
		]

		points_2 = [
			pos,
			[pos[0] / 2, (end[1] + pos[1]) / 2],
			end,
			[-self._offset, end[1] + self._offset],
		]

		res.extend(points_1)
		res.extend(points_2)

		t_points = np.arange(0, 1, 0.02)
		points1 = np.array(res)
		curve = Bezier.Curve(t_points, points1)

		vertices = []
		indices = []
		_i = 0

		for x, y in curve:
			vertices.extend([x, y, 0, 0])
			indices.append(_i)
			_i += 1

		self.indices = indices
		self.vertices = vertices

	def on_touch_move(self, touch):
		super().on_touch_move(touch)
		self.curve_pos = touch.pos

	def on_curve_pos(self, *args):
		self.update_mesh(self.start_pos[:], self.end_pos[:], self.curve_pos[:])

	def on_start_pos(self, *args):
		self.update_mesh(self.start_pos[:], self.end_pos[:], self.curve_pos[:])

	def on_touch_up(self, touch):
		super().on_touch_up(touch)

		distance = (touch.x - touch.ox)
		
		if abs(distance) > self.drag_distance:

			if distance > 0:
				start_pos = [self.width + self._offset, -self._offset]
				end_pos = [self.width + self._offset, self.height + self._offset]
				curve_pos = [self.width + self._offset, self.curve_pos[1]]
				self.move_anim("right", curve_pos, start_pos, end_pos)
				self.dispatch("on_right_start")
			else:
				start_pos = [0, 0]
				end_pos = [0, self.height]
				curve_pos = [0, self.curve_pos[1]]
				self.move_anim("left", curve_pos, start_pos, end_pos)
				self.dispatch("on_left_start")
		else:
			if self._state == "right":
				self.reset_anim([self.width + self._offset, touch.y])
				self.dispatch("on_reset_right")

			elif self._state == "left":
				self.reset_anim([0, touch.y])
				self.dispatch("on_reset_left")

	def move_anim(self, state, curve_pos, start_pos, end_pos):
		"""
		The curve has two animations. One for upper and lower corners
		and the other for the middle portion.
		"""
		anim = Animation(
			curve_pos=curve_pos,
			t="out_quad",
			d=0.4,
		)

		anim.bind(on_complete=lambda *args: self.dispatch(f"on_{state}_second_start"))
		
		anim2 = Animation(d=0.3) + Animation(
			start_pos=start_pos,
			end_pos=end_pos,
			t="in_quad",
			d=0.4,
		)

		anim2.bind(on_complete=lambda *args: self.dispatch(f"on_{state}_finish"))
		anim.start(self)
		anim2.start(self)
		self._state = state

	def reset_anim(self, curve_pos):
		anim = Animation(curve_pos=curve_pos, t="out_quad", d=0.2)
		anim.start(self)

	def add_widget(self, widget, **kw):
		if issubclass(widget.__class__, FrontLayout):
			self.ids.frontlayout.add_widget(widget)
			return True

		elif issubclass(widget.__class__, BackLayout):
			self.ids.backlayout.add_widget(widget)
			return True

		return super().add_widget(widget)




'''
[ Future ]:
	lbl = Label
	lbl.text ='Clique [b][ref=aqui]aqui[/ref][/b] para testar!'
	lbl.bind(on_ref_press=self.processaClique
	def processaClique(self,instance, value):
		print ("Usuario clicou sobre o link " , value)

[ IMPORTANTE ] | Adicionar seta de escrita quando estiver com focus == True |
[ BUGS ] Rolagem | Seleção | Canvas ViewPort |

'''

from kivy.app import runTouchApp
from kivy.core.text.markup import MarkupLabel as CoreLabel
from kivy.core.text import DEFAULT_FONT
from kivy.uix.widget import Widget
from kivy.core.window import Window
from kivy.properties import (
	ListProperty, NumericProperty,
	StringProperty, AliasProperty,
	BooleanProperty, ColorProperty
)
from kivy.clock import Clock
from kivy.metrics import dp, sp
from kivy.utils import get_color_from_hex
from kivy.graphics import Rectangle
from kivy.core.clipboard import Clipboard
from kivy.lang.builder import Builder
from functools import partial


meta_carac = {
	# specials keys
	8 : 'backspace',
	9 : 'tab',
	13 : 'enter',
	127 : 'delete',
	271 : 'enter',
	273 : 'up',
	274 : 'down',
	275 : 'right',
	276 : 'left',
	278 : 'home',
	279 : 'end',
	280 : 'pageup',
	281 : 'pagedown',
	301 : 'capslock',
	303 : 'lshift',
	304 : 'rshift',
	305 : 'lctrl',
	306 : 'rctrl',
	# 307 : 'alt-gr',
	307 : 'ralt',
	308 : 'lalt',
}


Builder.load_string("""


<CodeInput>:
	canvas.before:
		Color:
			rgba: self.background
		Rectangle:
			pos: self.pos
			size: self.size
		Color:
			rgba: self.text_color
	
	canvas.after:
		Color:
			rgba: self._cursor_color
		Rectangle:
			pos: self._cursor_pos
			size: self.cursor_size
		
		Color:
			rgba: self.selection_color[:-1] + [0.4]

""")


class CodeInput(Widget):
	
	cursor_color = ListProperty([[1, 0, 0, 1], [0, 1, 0, 1]])
	_cursor_color = ColorProperty([0, 0, 0, 0])
	
	cursor_size = ListProperty([sp(1), sp(18)])

	_cursor_pos = ListProperty([0, 0])
	cursor_interval = NumericProperty(0.6)
	_labels = ListProperty([])
	background = ColorProperty('#282a36')
	text_color = ColorProperty('#FFFFFF')
	selection_color = ColorProperty('#636672')

	tab_width = NumericProperty(4)

	font_name = StringProperty(DEFAULT_FONT)
	font_size = NumericProperty(sp(15))
	font_context = StringProperty(None, allownone=True)
	font_family = StringProperty(None, allownone=True)

	auto_indent = BooleanProperty(False)
	focus = BooleanProperty(False)


	# Index of the `end` & `init` visible labels on the viewport
	label_init = 0
	label_end = 0

	# Number of lines in all text -> len(lines)
	_n_labels = 0

	# Number of lines in the viewport -> viewport_height / line_height
	max_lines = 0
	
	# True if the key is pressed and False if not
	shift_pressed = False
	ctrl_pressed = False
	capslock_pressed = False
	alt_pressed = False

	# The position of the top line on viewport
	top_line_y = 0

	# Indicates if the input has a selection	
	in_selection = False

	# `Window.request_keyboard` to bind and get the normal keys
	_keyboard = None


	def _set_text(self, text:str) -> None:
		n_text = text.splitlines()
		self._labels.clear()
		self._rectangles.clear()
		self._n_labels = 0
		Clock.schedule_once(partial(self.insert_lines, n_text, True))

	def _get_text(self):
		return ''.join([f'{lbl.text}\n' for lbl in self._labels])
	
	text = AliasProperty(_get_text, _set_text, bind=('_labels',), cache=True)


	def __init__(self, **kwargs):
		self.selected = [0, 0]
		self.touch_down_pos = [[0, 0], [0, 0]]
		self.last_cursor_pos = [0, 0]
		self._cursor = [0, 0]
		self._selection = []
		self._rectangles = []
		
		super().__init__(**kwargs)
		self.cursor_clock = Clock.schedule_interval(self.change_cursor_color, self.cursor_interval)
		self.fbind('font_name', self.update_text_options)
		self.fbind('font_size', self.update_text_options)
		
		Window.bind(on_key_down=self.key_down, on_key_up=self.key_up)
		self.bind(size=self._update_graph, pos=self._update_graph)

		Clock.schedule_once(self.start)

	def start(self, *args) -> None:
		self._cursor_pos = [self.x, self.top_line_y]
		self.change_cursor_color("hide")

	def _update_graph(self, *args):
		self.update_values()
		self._update_lines()
	
	def _keyboard_unbind(self):
		if self._keyboard:
			self._keyboard.unbind(on_textinput=self.keyboard_on_textinput)
			self._keyboard.release()
			self._keyboard = None

	def _keyboard_bind(self):
		if not self._keyboard:
			self._keyboard = Window.request_keyboard(self._keyboard_unbind, self, 'text')
			self._keyboard.bind(on_textinput=self.keyboard_on_textinput)


	def change_cursor_color(self, dt, *args):
		color = self.get_color(self.cursor_color, 0)
		if self.in_selection:
			self._cursor_color = color
		
		elif self.focus:	
			if self._cursor_color != color or dt == "show":
				self._cursor_color = color
			else:
				self._cursor_color = self.get_color(self.cursor_color, 1, [*color[:-1], 0])
				
		self.cursor_clock.timeout = self.cursor_interval

	def update_text_options(self, *args):
		self.update_values()
		self._rectangles = []
		self._labels = []
		top_line_y = self.top_line_y

		for i_lbl in range(self._n_labels):
			label, texture, size = self.make_label(
											self._labels[i_lbl].text,
											self._label_kwargs)
			self._labels.append(label)
			self._rectangles.append(
				Rectangle(texture=texture, size=size, pos=[self.x, top_line_y])
			)
			top_line_y -= self.cursor_size[1]
		
		self.label_end = (self.label_init + self.max_lines)
		self._update_lines()

	def _update_lines(self, *args):
		if not self._labels:
			return None
		
		cur_y = self._cursor[1]
		cur_sh = self.cursor_size[1]
		new_max_lines = (self.label_init + self.max_lines)
		top_line_y = self.top_line_y

		if new_max_lines < self._n_labels:
			self.label_end = new_max_lines

		elif self.label_end < new_max_lines:
			if self.label_init > 0:
				self.label_init -= 1

			if self.label_end < self._n_labels:
				self.label_end += 1
		
		self.canvas.clear()
		for rect in self._rectangles[self.label_init:self.label_end]:
			rect.pos = [self.x, top_line_y]
			self.canvas.add(rect)
			top_line_y -= cur_sh
		
		y = self._rectangles[cur_y].pos[1]
		if self.label_init > cur_y:
			y = ( self.top_line_y + ( (self.label_init - cur_y) * cur_sh ) )

		elif self.label_end <= cur_y:
			y = ( top_line_y + cur_sh - ( (cur_y - self.label_end + 1) * cur_sh ) )

		self._cursor_pos = [self._x_cursor_by_index(*self._cursor), y]
		self.update_selection()

	def get_color(self, object, index, default=None):
		if isinstance(object, (list, tuple)):
			if len(object) == 2 and index > -1 and index < 3:
				color = object[index]
				if isinstance(color, str):
					color = get_color_from_hex(color)
				return color
		
		if index == 1 and default != None:
			return default
		return object

	def update_values(self, *args):
		self.cursor_size = [ dp(1), (1.2 * self.font_size) ]
		cur_sh = self.cursor_size[1]
		self.max_lines = int(round((self.height / cur_sh), 0))
		self.top_line_y = (self.y + self.height - cur_sh)

	def move_cursor_left(self, cur_sw, cur_sh, cur_x, cur_y, cur_px, cur_py):

		if (self.x + cur_sw) >= cur_px and cur_y > 0:
			if cur_y == self.label_init:
				self._scroll_down()
				cur_py -= cur_sh
			
			cur_y -= 1
			cur_x = self.get_len_str(cur_y)
			cur_px = (self.x + cur_sw + self._rectangles[cur_y].size[0])
			cur_py += cur_sh
		
		elif cur_px > (self.x + cur_sw):
			if self.ctrl_pressed:
				cur_px, cur_x = self.last_word()
			else:
				cur_x -= 1
				cur_px = self._x_cursor_by_index(cur_x, cur_y)
		
		self.ctrl_selection([cur_x, cur_y], [cur_px, cur_py])
	
	def move_cursor_right(self, largs):
		cur_sw, cur_sh, cur_x, cur_y, cur_px, cur_py = largs

		n_lbl = self._n_labels-1
		if self._rectangles:
			size_rec = (self._rectangles[cur_y].size[0] + self.x)
			if size_rec <= cur_px and cur_y < n_lbl:
				cur_y += 1
				cur_x = 0
				cur_px, cur_py = self.x, cur_py-cur_sh
				
			elif cur_y <= n_lbl and cur_px <= size_rec:
				if size_rec == 1:
					if cur_y < n_lbl:
						cur_y += 1
						cur_py -= cur_sh
				elif self.ctrl_pressed:
					cur_px, cur_x = self.next_word(cur_y)
				else:
					cur_x += 1
					get_pos = self._x_cursor_by_index(cur_x, cur_y)
					if get_pos != cur_px:
						cur_px = get_pos
					else:
						cur_x -= 1
			
			if cur_y == self.label_end:
				self._scroll_up()
				cur_py += cur_sh

		self.ctrl_selection([cur_x, cur_y], [cur_px, cur_py])
	
	def move_cursor_up(self, largs):
		cur_sw, cur_sh, cur_x, cur_y, cur_px, cur_py = largs

		if cur_y > 0:
			if cur_y == self.label_init:
				self._scroll_down()
			else:
				cur_py += cur_sh
			cur_y -= 1
			cur_px = (self._x_cursor_by_index(cur_x, cur_y) + self.x)
			if cur_x >= self.get_len_str(cur_y):
				cur_x = self.get_len_str(cur_y)

		self.ctrl_selection([cur_x, cur_y], [cur_px, cur_py])
	
	def move_cursor_down(self, largs):
		cur_sw, cur_sh, cur_x, cur_y, cur_px, cur_py = largs

		if cur_y < self._n_labels-1:
			if cur_y == self.label_end-1:
				self._scroll_up()
			else:
				cur_py -= cur_sh
			cur_y += 1
			cur_px = (self._x_cursor_by_index(cur_x, cur_y) + self.x)
			if cur_x >= self.get_len_str(cur_y):
				cur_x = self.get_len_str(cur_y)
				
		self.ctrl_selection([cur_x, cur_y], [cur_px, cur_py])

	def ctrl_selection(self, cursor, cursor_pos):
		self._cursor = [*cursor]
		
		if self.ctrl_pressed:
			self.last_cursor_pos = [*cursor_pos]
			self.do_selection()

		self._cursor_pos = [*cursor_pos]

	def update_selection(self):
		self.canvas.after.remove_group('selection')
		
		for i in self._selection:
			rect, index = i[0], i[1][0]
			if index >= self.label_init and index < self.label_end:
				rect.pos = [rect.pos[0], self._rectangles[index].pos[1]]
				rect.size = [rect.size[0], self.cursor_size[1]]
				self.canvas.after.add(rect)

	def do_selection(self, *args):
		if not self._labels:
			return None
		

		last_cursor_pos, last_cursor = self.get_cursor_by_touch(*self.last_cursor_pos, False)
		_cursor_pos, _cursor = self.touch_down_pos
		
		cur_y = self._cursor[1]
		largs = [
			last_cursor[1], _cursor[1],
			last_cursor_pos[0], last_cursor[0],
			_cursor_pos[0], _cursor[0]
		]

		if not self._selection:
			rect = Rectangle(size=[0, 0], pos=[0, 0], group='selection')
			self._selection.append([rect, [_cursor[1], self.x]])
			self.selected[1] += 1
			self.do_selection()
			self.selected[0] = cur_y
			return None
		
		self.remove_selection(False)
		l_select = max(_cursor[1], cur_y)
		first_index = min(_cursor[1], cur_y)
		total_select = len(self._selection)-1
		
		if total_select > 0:
			if l_select == self.max_lines-1:
				l_select = self.label_end-1
			if first_index == self.max_lines-1-self.label_init:
				first_index = self.label_init

		total_indexs = (l_select - first_index)

		for n, index in enumerate(range(first_index, l_select+1)):

			if index < self.label_init or index > self.label_end:
				continue

			lbl = self._labels[index]
			width, pos_x = lbl.size[0], self.x
			
			if total_indexs == 0:
				width, pos_x = self.direction_first_select(lbl, largs)
			
			elif index == first_index:
				x, px = self.direction_select(largs, last_select=False)
				width = lbl.get_extents(lbl.text[x:])[0]
				pos_x = px
			
			elif index == l_select:
				x, px = self.direction_select(largs, last_select=True)
				width = lbl.get_extents(lbl.text[:x])[0]

			try:
				if index == (self.label_end - 1) or index == (self.label_init + 1):
					rect = Rectangle(
								size=[width, self.cursor_size[1]],
								pos=[pos_x, self._rectangles[index].pos[1]],
								group='selection'
					)
					self._selection.append([rect, [index, pos_x]])
					self.selected[1] += 1
				else:
					rect = self._selection[n][0]
					self._selection[n] = [rect, [index, pos_x]]
					rect.size = [width, self.cursor_size[1]]
					rect.pos = [pos_x, self._rectangles[index].pos[1]]
			except IndexError:
				rect = Rectangle(
							size=[width, self.cursor_size[1]],
							pos=[pos_x, self._rectangles[index].pos[1]],
							group='selection'
				)
				self._selection.append([rect, [index, pos_x]])
				self.selected[1] += 1
			
			self.canvas.after.add(rect)
		
		self._selection = self._selection[:total_indexs+1]
		self.selected[0] = (cur_y - 1)

	def direction_first_select(self, lbl, largs):
		last_y, touch_y, last_px, last_x, touch_px, touch_x = largs
		
		if last_x < touch_x:
			width = lbl.get_extents(lbl.text[last_x:touch_x])[0]
			pos_x = last_px
		else:
			width = lbl.get_extents(lbl.text[touch_x:last_x])[0]
			pos_x = touch_px

		return [width, pos_x]

	def direction_select(self, largs, last_select):
		last_y, touch_y, last_px, last_x, touch_px, touch_x = largs
		if last_select:
			if last_y > touch_y:
				return [last_x, self.x]
			else:
				return [touch_x, self.x]
		else:
			if last_y < touch_y:
				return [last_x, last_px]
			else:
				return [touch_x, touch_px]
			
	def remove_selection(self, remov=True):
		self.canvas.after.remove_group('selection')
		if remov:
			self._selection = []
			self.touch_down_pos = [[0, 0], [0, 0]]
			self.last_cursor_pos = [0, 0]
			self.selected = [0, 0]

	def get_len_str(self, cur_y):
		if not self._labels:
			return 0
		
		return len(self._labels[cur_y].text)

	def _x_cursor_by_index(self, cur_x, cur_y):
		if not self._labels:
			return self.x
		
		lbl = self._labels[cur_y]
		width = lbl.get_extents(lbl.text[:cur_x])[0]
		return (self.x + width) 
	
	def insert_lines(self, lines, set_text=False, *args):
		if not lines:
			return None
	
		lbl_txt = ''
		lbls = []
		recs = []
		cur_sh = self.cursor_size[1]
		cur_py = self._cursor_pos[1]
		cur_x, cur_y = self._cursor
		self.canvas.clear()

		if self._labels:
			lbl_txt = self._labels[cur_y].text[:cur_x]
			lines[-1] += self._labels[cur_y].text[cur_x:]
		
		l_lines = len(lines)
		
		for n in range(l_lines):
			text = ((lbl_txt if n == 0 else '') + lines[n])
			label, texture, size = self.make_label(text, self._label_kwargs)
			lbls.append(label)
			recs.append(Rectangle(texture=texture, size=size,
								  pos=[self.x, cur_py]))
			cur_py -= cur_sh
		
		self._n_labels += (l_lines - 1) if self._labels else (l_lines)
		self._rectangles = (self._rectangles[:cur_y] + recs + self._rectangles[cur_y+1:])
		self._labels = (self._labels[:cur_y] + lbls + self._labels[cur_y+1:])
		
		if self._n_labels <= self.max_lines:
			self.label_end = (self.label_init + self._n_labels)
			self._cursor[1] += (l_lines - 1)
			self._cursor[0] = len(lbls[-1].text)
		else:
			self.label_init = (self.label_end - self.max_lines)
			self.label_end = (cur_y + l_lines)
			self._cursor[0] = [len(lbls[-1].text), (self.label_end - 1)]

			if self.label_init < 0:
				self._cursor[1] = self.label_end
				self.label_init = 0
				self.label_end = self.max_lines
		
		if set_text:
			self._cursor = [0, 0]
			self.label_init = 0
			self.label_end = l_lines

			if l_lines > self.max_lines:
				self.label_end = self.max_lines
				
		self._update_lines()

	def ctrl_v(self):
		lines = Clipboard.paste().splitlines()
		self.insert_lines(lines)

	def ctrl_c(self):
		pass

	def ctrl_z(self):
		pass

	def ctrl_x(self):
		pass

	def do_redo(self):
		pass

	def backspace(self):
		cur_sw, cur_sh = self.cursor_size
		cur_px, cur_py = self._cursor_pos
		x_before, y_before = self._cursor
		
		self.move_cursor_left(cur_sw, cur_sh, *self._cursor, cur_px, cur_py)
		
		cur_x, cur_y = self._cursor
		lbls, recs = self._labels, self._rectangles
		
		if lbls and x_before > 0:
			t_label = lbls[y_before].text
			n_text = t_label[:cur_x] + t_label[x_before:]
			self.change_line(n_text, y_before, self._label_kwargs)
		
		elif y_before > 0:
			n_text = lbls[cur_y].text[:cur_x] + lbls[y_before].text[x_before:]
			self.change_line(n_text, cur_y, self._label_kwargs)
			
			del lbls[y_before]
			del recs[y_before]
			self._n_labels -= 1
			self.label_end -= 1
			
			if self.label_end == self._n_labels and self.label_init > 0:
				self.label_init -= 1
			
			self._update_lines()

	def enter(self):
		cur_x, cur_y = self._cursor
		text = self._labels[cur_y].text
		self._cursor = [0, (cur_y + 1)]
		n_text = ''

		if self.auto_indent:
			tabs = self.identation(0, cur_y)
			n_text = ' ' * tabs
			self._cursor[0] = tabs

		if text[cur_x:] == '':
			cur = (-1) if self._cursor[1] > self._n_labels else (self._cursor[1])
			self.insert_string(n_text, cur)
		else:
			self.change_line(text[:cur_x], cur_y, self._label_kwargs)
			self.insert_string((n_text + text[cur_x:]), self._cursor[1])
		
		self._update_lines()
	
	def identation(self, cur_x, cur_y):
		for n, x in enumerate(self._labels[cur_y].text):
			if x != ' ':
				return n
		return cur_x

	def next_word(self, cur_y):
		label  = self._labels[cur_y]
		text = label.text

		for n, i in enumerate(text):
			if n > self._cursor[0] and i == ' ':
				w_text = label.get_extents(text[:n])[0]
				return [(self.x + w_text), len(text[:n])]
		
		line_height = (self.x + self._rectangles[cur_y].size[0])
		return [(self.x + line_height), self.get_len_str(cur_y)]
	
	def last_word(self):
		label  = self._labels[self._cursor[1]]
		get_ext = label.get_extents
		text = label.text
		l_txt = local = len(text)

		for n, i in enumerate(label.text[::-1]):
			local = (l_txt - n - 1)
			if self._cursor[0] > local and i == ' ':
				lbl_w = get_ext(text[:local])[0]
				return [(self.x + lbl_w), len(text[:local])]
		
		lbl_w = get_ext(text[:local])[0]
		return [(self.x + lbl_w), len(text[:local])]
		
	
	def _scroll_up(self):
		cur_sh = self.cursor_size[1]
		recs = self._rectangles
		pos_i = (self.y + self.height - cur_sh)
		
		if self.label_end >= self._n_labels:
			return None
		
		self.canvas.remove(recs[self.label_init])
		
		for rect in recs[ (self.label_init + 1) : (self.label_end + 1) ]:
			rect.pos = [self.x, pos_i]
			pos_i -= cur_sh

		self.canvas.add(recs[self.label_end])

		self.label_init += 1
		self.label_end += 1
		self._cursor_pos[1] += cur_sh
		self.update_selection()

		if self.in_selection:
			self.do_selection()

	def _scroll_down(self):
		if self.label_init == 0:
			return None
		
		cur_sh = self.cursor_size[1]
		recs = self._rectangles
		pos_i = (self.y + self.height - cur_sh)
	
		rect1 = recs[self.label_init-1]
		self.canvas.remove(recs[self.label_end-1])

		for rect in recs[self.label_init:self.label_end-1]:
			pos_i -= cur_sh
			rect.pos = [self.x, pos_i]

		rect1.pos = [ self.x, (self.y + self.height - self.cursor_size[1]) ]
		self.canvas.add(rect1)

		self.label_init -= 1
		self.label_end -= 1
		self._cursor_pos[1] -= cur_sh
		self.update_selection()

		if self.in_selection:
			self.do_selection()

	def key_up(self, window, key, *args):
		if not self.focus:
			return False

		chr_key = ( meta_carac[key] ) if key in meta_carac else ( chr(key) )
		
		if chr_key in {'rshift', 'lshift'}:
			self.shift_pressed = False
		
		elif chr_key in {'rctrl', 'lctrl'}:
			self.ctrl_pressed = False

		elif chr_key in {'ralt', 'lalt', 'alt-gr'}:
			self.alt_pressed = False
		
		return True

	def key_down(self, window, key, *args):
		if not self.focus:
			return False

		chr_key = ( meta_carac[key] ) if key in meta_carac else ( chr(key) )
		
		if chr_key in ['backspace', 'delete']:
			self.backspace()

		elif chr_key == 'enter':
			self.enter()

		elif self.ctrl_pressed and self.shift_pressed and chr_key == 'z':
			self.do_redo()

		elif self.ctrl_pressed and chr_key in ['v', 'c', 'z', 'x']:
			if self.focus:
				func = getattr(self, f'ctrl_{chr_key}')
				func()
		
		elif chr_key in ['up', 'down', 'left', 'right']:
			move_cursor = getattr(self, f'move_cursor_{chr_key}')
			move_cursor([*self.cursor_size, *self._cursor, *self._cursor_pos])
		
		elif chr_key in ['rshift', 'lshift', 'rctrl', 'lctrl']:
			setattr(self, chr_key[1:], True)
		
		elif chr_key in {'ralt', 'lalt', 'alt-gr'}:
			setattr(self, chr_key, True)
		
		elif chr_key == 'capslock':
			self.capslock_pressed = not self.capslock_pressed
		
		elif chr_key in ['tab']:
			self.insert_string(u'\t')
		
		elif chr_key in ['pageup', 'pagedown', 'end', 'home']:
			pass

		self.cursor_clock.timeout = (self.cursor_interval + 1.4)
		self.change_cursor_color("show")
		return True
	
	def keyboard_on_textinput(self, window, chr_key):
		if not self.focus:
			return False
		
		cur_sw, cur_sh = self.cursor_size 

		if self._cursor_pos[0] > (self.x + self.width):
			self._cursor_pos = [ (self.x + cur_sw), (self._cursor_pos[1] - cur_sh) ]
			self._cursor[1] += 1
			self._cursor[0] = 0
		
		self.insert_string(chr_key)

	def _scroll_to(self, cur_y):
		if not self._labels:
			return None
		
		if self.label_init < cur_y:
			lbl_init = (cur_y - self.max_lines + 1)
			lbl_end = (cur_y + 1)
			cur_p = (lbl_end - 1)
		
		elif self.label_init > cur_y:
			lbl_init = cur_y
			lbl_end = (cur_y + self.max_lines)
			cur_p = lbl_init

		self.label_init = lbl_init
		self.label_end = lbl_end
		self._cursor_pos[1] = self._rectangles[cur_p].pos[1]
		self._cursor[1] = cur_y
		self._update_lines()

	@property
	def _label_kwargs(self):
		return {
			'font_size': self.font_size,
			'font_name': self.font_name,
			'font_context': self.font_context,
			'font_family': self.font_family,
			'markup': True,
		}

	def change_line(self, n_text, y_cur, kw):
		label, texture, size = self.make_label(n_text, kw)
		self._labels[y_cur] = label

		self.canvas.remove(self._rectangles[y_cur])
		self._rectangles[y_cur] = Rectangle(texture=texture, size=size,
											pos=[self.x, self._cursor_pos[1]])
		self.canvas.add(self._rectangles[y_cur])

	def make_label(self, text, kwargs):
		n_text = text.replace(u'[', u'&bl;').replace(u']', u'&br;')
		# n_text = ''.join((u'[color=#FF0000]', n_text, u'[/color]'))
		label = CoreLabel(text=n_text, **kwargs)
		label.refresh()
		label.text = text
		return [label, label.texture, label.texture.size]

	def insert_string(self, string, p=-1):
		cur_x, cur_y = self._cursor
		cur_sh = self.cursor_size[1]
		new_cur_y = (self._cursor_pos[1] + cur_sh)

		if self.capslock_pressed or self.shift_pressed and p != -1:
			string = string.upper()

		if string.find(u'\t') != -1:
			string = string.replace(u'\t', u' '*self.tab_width)
			self._cursor[0] += (len(string) - 1)

		if self._n_labels > 0 and self._n_labels > cur_y and p == -1:
			text = self._labels[cur_y].text
			n_text = text[:cur_x] + string + text[cur_x:]
			self.change_line(n_text, cur_y, self._label_kwargs)
		else:
			label, texture, size = self.make_label(string, self._label_kwargs)
			self._labels.insert(p, label)
			self._n_labels += 1

			if self._n_labels <= self.max_lines:
				self.label_end = self._n_labels

			rect = Rectangle(texture=texture, size=size,
		    				 pos=[self.x, self._cursor_pos[1]])
			self._rectangles.insert(p, rect)
			self.canvas.add(rect)

		if p == -1:
			self._cursor[0] += 1
		
		self._cursor_pos[0] = self._x_cursor_by_index(*self._cursor)

		if self._rectangles[self._cursor[1]].pos[1] < self.y or p != -1:
			self._scroll_up()
		
		if string == '' or p != -1:
			self._cursor_pos[1] -= self.cursor_size[1]
		
		if new_cur_y > ( self.y + self.height + cur_sh ) or \
			new_cur_y < ( self.y + (cur_sh * 2) ):
			self._scroll_to(cur_y)

	def change_focus(self, in_focus):
		if in_focus:
			return None
		
		self.focus = in_focus
		self.change_cursor_color("hide")
		self._keyboard_unbind()

	def on_touch_up(self, touch):
		if self.in_selection:
			self.in_selection = False
			self.change_cursor_color("show")
			self.cursor_clock.cancel()
			self.cursor_clock()

		if not self.collide_point(*touch.pos):
			self.change_focus(False)
			return False

	def on_touch_down(self, touch):
		if not self.collide_point(*touch.pos):
			self.change_focus(False)
			return False

		if super().on_touch_down(touch):
			return True

		if self._keyboard == None:
			self._keyboard_bind()

		self.cursor_clock.timeout = (self.cursor_interval + 1.4)
		self.change_cursor_color("show")

		if 'button' in touch.profile and touch.button.startswith('scroll'):
			scroll_type = touch.button[6:]

			if scroll_type == 'down':
				self._scroll_down()
			
			elif scroll_type == 'up':
				self._scroll_up()
		
		elif self.focus:
			self.in_selection = False
			self.remove_selection()
			
			self.get_cursor_by_touch(*touch.pos)
			self.last_cursor_pos = [*self._cursor_pos]
			self.touch_down_pos = [[*self._cursor_pos], [*self._cursor]]
			self.selected = [self._cursor[1], 0]

		self.focus = True
		return True
	
	def on_touch_move(self, touch):
		if not self.focus:
			return None
		
		print("touch.pos -=> ", touch.pos)
		self.get_cursor_by_touch(*touch.pos)
		self.last_cursor_pos = [*self._cursor_pos]
		self.in_selection = True

		if touch.pos[1] <= ( self.y + (self.cursor_size[1] / 2) ):
			self._scroll_up()
		
		elif touch.pos[1] >= ( self.y + self.height - (self.cursor_size[1] / 2) ):
			self._scroll_down()
		
		else:
			self.do_selection()
		

	def get_cursor_by_touch(self, x, y, change=True):
		if not self._labels:
			return None
		
		cur_sh = self.cursor_size[1]
		cur_y = int(round( (( (y - self.y) / cur_sh ) + 0.3), 0 ))
		cur_y = (self.max_lines - cur_y + self.label_init)

		if cur_y >= self._n_labels:
			return None
		
		lbl = self._labels[cur_y]
		l_txt = len(lbl.text)
		width = lbl.size[0]
		divisor = ( max(width, 1) / max(l_txt, 1) )
		cur_x = min( int(round( ((x - self.x) / divisor), 0 )), l_txt)
		
		cur_p = ( self.y + self.height - ( (cur_y - self.label_init) * cur_sh ) - cur_sh )
		if cur_p < self.y:
			cur_p = self._cursor_pos[1]
		
		pos = (self.x + lbl.get_extents(lbl.text[:cur_x])[0])
		if change:
			self._cursor_pos = [pos, cur_p]
			self._cursor = [cur_x, cur_y]

		print("by -=> ", x, y)
		print("position -=> ", [[pos, cur_p], [cur_x, cur_y]])
		return [[pos, cur_p], [cur_x, cur_y]]



if __name__ == '__main__':

	path = r'C:\Users\heito\AppData\Local\Programs\Python\Python39\Lib\site-packages\kivy\uix\textinput.py'
	from kivy.uix.textinput import TextInput

	with open(path, 'r') as file:
		text = ""
		text = file.read()
		# inp = TextInput()
		inp = CodeInput()
		inp.text = text
		runTouchApp(inp)



from kivy.clock import Clock
from kivy.lang import Builder
from kivy.properties import (
	ColorProperty, ListProperty,
	NumericProperty, ObjectProperty
)
from kivy.uix.slider import Slider
from kivygo.behaviors.neumorph import NeuGlowCircular
from kivygo.behaviors.thumb import Thumb
from kivy.metrics import dp
from kivy.graphics.texture import Texture
from kivygo.utils import dec_2_rgb
from PIL import Image, ImageDraw, ImageFilter


Builder.load_string("""

<NeuSlider>
	canvas:
		Clear
		Color:
			rgba: self.background_color
		RoundedRectangle:
			size: self.size
			pos: self.pos
			radius: self.radius

	background_width: 0
	cursor_size: [0, 0]
	padding: (self.height / 2)

	NeuThumb:
		canvas.before:
			Color:
			Rectangle:
				size:self.glow_size
				pos:self.glow_pos
				texture:self.glow_texture
		canvas:
			Color:
				rgba:root.thumb_color
			Ellipse:
				size: [ (root.height - root.thumb_padding), (root.height - root.thumb_padding) ]
				pos:self.pos
			Color:

		size_hint: [None, None]
		size: [( root.height - root.thumb_padding ), ( root.height - root.thumb_padding ) ]
		pos: [ ( root.value_pos[0] - self.width / 2), ( root.center_y - self.height / 2) ]
		behind_color: ( root.comp_color[0:3] + [0] ) if root.thumb_bg_color == [0, 0, 0, 0] else root.thumb_bg_color
		glow_color: root.thumb_color if root.glow_color == [0, 0, 0, 0] else root.glow_color

""")


class NeuSlider(Slider):

	comp_color = ColorProperty("#333333")

	shadow_behind_color = ColorProperty("#333333")

	background_color = ListProperty([0.6, 0.1, 0.4, 1])

	elevation = NumericProperty(dp(3))
	"""
	Elevation of the widget.Elevation can be any number between -5 and +5(inclusive).
	Negative elevation will cause the widget to go into the screen whereas positive
	elevation will make it pop from the screen
	"""

	radius = ListProperty([dp(20), dp(20), dp(20), dp(20)])
	"""
	Radius of the slider bar. The value defaults to half the height of the slider bar.
	"""

	thumb_color = ColorProperty([0, 0, 0, 0])
	"""
	Color of the thumb of the slider
	"""

	thumb_bg_color = ColorProperty([0, 0, 0, 0])
	"""
	Color of background behind the thumb. This property is needed to
	properly display the glow effect. The property will default to the component
	color. But it can be manually set. If a color is manually set it will create a
	ring of that color around the thumb's glow.

	attr:`thumb_bg_color` is an :class:`~kivy.properties.ColorProperty`
	and defaults to the slider's 'comp_color'.
	"""

	thumb_padding = NumericProperty(0)
	"""
	The top and bottom padding value for the the thumb. This allows you to inset
	the thumb in the slider. It defaults to zero which means the thumb will be
	as big as the height of the slider.
	"""

	glow_color = ColorProperty([0, 0, 0, 0])
	"""
	Color of the glow behind the thumb. Defaults to the thumb's color.
	"""

	glow_radius = NumericProperty(dp(20))
	"""
	Radius of the glow behind the thumb.
	"""

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		Clock.schedule_once(self.elevation_set)
		Clock.schedule_once(self.radius_set)

	def elevation_set(self, *args):
		self.elev = self.elevation

	def radius_set(self, *args):
		self.radius = [(self.height / 2)] * 4


class NeuThumb(Thumb, NeuGlowCircular):

	glow_color = ColorProperty([0.8, 0.7, 0.5, 1])

	elevation = NumericProperty(0)

	glow_radius = NumericProperty(dp(20))
	"""
	Radius of the glow
	"""

	color = ColorProperty([0, 0, 0, 0])
	"""
	Color of the glow. It is suggested to set this property to the color of the
	component that the glow is being added to
	"""

	behind_color = ColorProperty([0, 0, 0, 0])
	"""
	Color of the component behind the glow. If left blank it will result in
	a black ring around your glow, so make sure you set this color properly
	"""

	glow_texture = ObjectProperty()
	"""
	The property that holds the actual glow texture object
	"""

	glow_size = ListProperty([0, 0])
	"""
	A list containing the size of the glow texture
	"""

	glow_pos = ListProperty([0, 0])
	"""
	A list containing the position of the glow texture
	"""


	def on_size(self, *args):
		self._create_glow()

	def on_pos(self, *args):
		self._update_glow_pos()

	def on_glow_radius(self, *args):
		self._create_glow()


	def __init__(self, **kwargs):
		self.elev = self.elevation
		self.blank_texture = Texture.create(size=self.size, colorfmt="rgba")
		Clock.schedule_once(self._create_glow)
		super().__init__(**kwargs)


	def _create_glow(self, *args):
		# Create blank image
		blank_x_size = int(self.size[0] + self.glow_radius * 2)
		blank_y_size = int(self.size[1] + self.glow_radius * 2)
		
		shadow = Image.new(
			"RGBA",
			(blank_x_size, blank_y_size),
			color=(tuple(dec_2_rgb(self.behind_color))),
		)

		# Convert to drawable image
		blank_draw = ImageDraw.Draw(shadow)
		
		x0 = (blank_x_size - self.size[0]) / 2.0
		y0 = (blank_y_size - self.size[1]) / 2.0
		
		x1 = (x0 + self.size[0])
		y1 = (y0 + self.size[1])

		blank_draw.ellipse(
			[(x0, y0), (x1, y1)],
			fill=tuple(dec_2_rgb(self.glow_color)),
		)

		texture = Texture.create(size=(blank_x_size, blank_y_size), colorfmt="rgba")
		shadow = shadow.filter(ImageFilter.GaussianBlur(self.glow_radius / 2))
		texture.blit_buffer(shadow.tobytes(), colorfmt="rgba", bufferfmt="ubyte")
		
		self.glow_texture = texture
		self.glow_size = (blank_x_size, blank_y_size)
		self.glow_pos = [(self.pos[0] - x0), (self.pos[1] - y0)]

	def _update_glow_pos(self):
		self.glow_pos = [(self.pos[0] - self.glow_radius), (self.pos[1] - self.glow_radius)]





from math import sqrt, cos, sin, pi
from collections import ChainMap
from kivy.event import EventDispatcher
from kivy.clock import Clock
from kivy.compat import string_types, iterkeys
from kivy.weakproxy import WeakProxy


class Animation(EventDispatcher):
    '''Create an animation definition that can be used to animate a Widget.
    :Parameters:
        `duration` or `d`: float, defaults to 1.
            Duration of the animation, in seconds.
        `transition` or `t`: str or func
            Transition function for animate properties. It can be the name of a
            method from :class:`AnimationTransition`.
        `step` or `s`: float
            Step in milliseconds of the animation. Defaults to 0, which means
            the animation is updated for every frame.
            To update the animation less often, set the step value to a float.
            For example, if you want to animate at 30 FPS, use s=1/30.
    :Events:
        `on_start`: animation, widget
            Fired when the animation is started on a widget.
        `on_complete`: animation, widget
            Fired when the animation is completed or stopped on a widget.
        `on_progress`: animation, widget, progression
            Fired when the progression of the animation is changing.
    .. versionchanged:: 1.4.0
        Added s/step parameter.
    .. versionchanged:: 1.10.0
        The default value of the step parameter was changed from 1/60. to 0.
    '''

    _update_ev = None

    _instances = set()

    __events__ = ('on_start', 'on_progress', 'on_complete')


    def on_start(self, widget):
        pass

    def on_progress(self, widget, progress):
        pass

    def on_complete(self, widget):
        pass


    def __init__(self, **kw):
        super().__init__()
        # Initialize
        self._clock_installed = False
        self._duration = kw.pop('d', kw.pop('duration', 1.))
        self._transition = kw.pop('t', kw.pop('transition', 'linear'))
        self._step = kw.pop('s', kw.pop('step', 0))
        if isinstance(self._transition, string_types):
            self._transition = getattr(AnimationTransition, self._transition)
        self._animated_properties = kw
        self._widgets = {}

    @property
    def duration(self):
        '''Return the duration of the animation.
        '''
        return self._duration

    @property
    def transition(self):
        '''Return the transition of the animation.
        '''
        return self._transition

    @property
    def animated_properties(self):
        '''Return the properties used to animate.
        '''
        return self._animated_properties

    @staticmethod
    def stop_all(widget, *largs):
        '''Stop all animations that concern a specific widget / list of
        properties.
        Example::
            anim = Animation(x=50)
            anim.start(widget)
            # and later
            Animation.stop_all(widget, 'x')
        '''
        if len(largs):
            for animation in list(Animation._instances):
                for x in largs:
                    animation.stop_property(widget, x)
        else:
            for animation in set(Animation._instances):
                animation.stop(widget)

    @staticmethod
    def cancel_all(widget, *largs):
        '''Cancel all animations that concern a specific widget / list of
        properties. See :attr:`cancel`.
        Example::
            anim = Animation(x=50)
            anim.start(widget)
            # and later
            Animation.cancel_all(widget, 'x')
        .. versionadded:: 1.4.0
        .. versionchanged:: 2.1.0
            If the parameter ``widget`` == None, all animated widgets will be
            the target and cancelled. If ``largs`` is also given, animation of
            these properties will be canceled for all animated widgets.
        '''
        if widget == None:
            if largs:
                for animation in Animation._instances.copy():
                    for info in tuple(animation._widgets.values()):
                        widget = info['widget']
                        for x in largs:
                            animation.cancel_property(widget, x)
            else:
                for animation in Animation._instances:
                    animation._widgets.clear()
                    animation._clock_uninstall()
                Animation._instances.clear()
            return None
        if len(largs):
            for animation in list(Animation._instances):
                for x in largs:
                    animation.cancel_property(widget, x)
        else:
            for animation in set(Animation._instances):
                animation.cancel(widget)

    def start(self, widget):
        '''Start the animation on a widget.
        '''
        self.stop(widget)
        self._initialize(widget)
        self._register()
        self.dispatch('on_start', widget)

    def stop(self, widget):
        '''Stop the animation previously applied to a widget, triggering the
        `on_complete` event.'''
        props = self._widgets.pop(widget.uid, None)
        if props:
            self.dispatch('on_complete', widget)
        self.cancel(widget)

    def cancel(self, widget):
        '''Cancel the animation previously applied to a widget. Same
        effect as :attr:`stop`, except the `on_complete` event will
        *not* be triggered!
        .. versionadded:: 1.4.0
        '''
        self._widgets.pop(widget.uid, None)
        self._clock_uninstall()
        if not self._widgets:
            self._unregister()

    def stop_property(self, widget, prop):
        '''Even if an animation is running, remove a property. It will not be
        animated further. If it was the only/last property being animated,
        the animation will be stopped (see :attr:`stop`).
        '''
        props = self._widgets.get(widget.uid, None)
        if not props:
            return None
        props['properties'].pop(prop, None)

        # no more properties to animation ? kill the animation.
        if not props['properties']:
            self.stop(widget)

    def cancel_property(self, widget, prop):
        '''Even if an animation is running, remove a property. It will not be
        animated further. If it was the only/last property being animated,
        the animation will be canceled (see :attr:`cancel`)
        .. versionadded:: 1.4.0
        '''
        props = self._widgets.get(widget.uid, None)
        if not props:
            return None
        props['properties'].pop(prop, None)

        # no more properties to animation ? kill the animation.
        if not props['properties']:
            self.cancel(widget)

    def have_properties_to_animate(self, widget):
        '''Return True if a widget still has properties to animate.
        .. versionadded:: 1.8.0
        '''
        props = self._widgets.get(widget.uid, None)
        if props and props['properties']:
            return True

    #
    # Private
    #
    def _register(self):
        Animation._instances.add(self)

    def _unregister(self):
        if self in Animation._instances:
            Animation._instances.remove(self)

    def _initialize(self, widget):
        d = self._widgets[widget.uid] = {
            'widget': widget,
            'properties': {},
            'time': None}

        # get current values
        p = d['properties']
        for key, value in self._animated_properties.items():
            original_value = getattr(widget, key)
            if isinstance(original_value, (tuple, list)):
                original_value = original_value[:]
            elif isinstance(original_value, dict):
                original_value = original_value.copy()
            p[key] = (original_value, value)

        # install clock
        self._clock_install()

    def _clock_install(self):
        if self._clock_installed:
            return None
        
        self._update_ev = Clock.schedule_interval(self._update, self._step)
        self._clock_installed = True

    def _clock_uninstall(self):
        if self._widgets or not self._clock_installed:
            return None
        
        self._clock_installed = False
        if self._update_ev != None:
            self._update_ev.cancel()
            self._update_ev = None

    def _update(self, dt):
        widgets = self._widgets
        transition = self._transition
        calculate = self._calculate
        for uid in list(widgets.keys()):
            anim = widgets[uid]
            widget = anim['widget']

            if isinstance(widget, WeakProxy) and not len(dir(widget)):
                # empty proxy, widget is gone. ref: #2458
                self._widgets.pop(uid, None)
                self._clock_uninstall()
                
                if not self._widgets:
                    self._unregister()
                
                continue

            if anim['time'] == None:
                anim['time'] = 0.0
            else:
                anim['time'] += dt

            # calculate progression
            if self._duration:
                progress = min(1.0, anim['time'] / self._duration)
            else:
                progress = 1
            t = transition(progress)

            # apply progression on widget
            for key, values in anim['properties'].items():
                a, b = values
                value = calculate(a, b, t)
                setattr(widget, key, value)

            self.dispatch('on_progress', widget, progress)

            # time to stop ?
            if progress >= 1.0:
                self.stop(widget)

    def _calculate(self, a, b, t):
        _calculate = self._calculate
        
        if isinstance(a, list) or isinstance(a, tuple):
            
            if isinstance(a, list):
                tp = list
            else:
                tp = tuple

            return tp([_calculate(a[x], b[x], t) for x in range(len(a))])
        
        elif isinstance(a, dict):
            
            d = {}
            for x in iterkeys(a):
                if x not in b:
                    # User requested to animate only part of the dict.
                    # Copy the rest
                    d[x] = a[x]
                else:
                    d[x] = _calculate(a[x], b[x], t)
            return d
        else:
            return (a * (1.0 - t)) + (b * t)

    #
    # Default handlers
    #

    def __add__(self, animation):
        return Sequence(self, animation)

    def __and__(self, animation):
        return Parallel(self, animation)


class CompoundAnimation(Animation):

    def stop_property(self, widget, prop):
        self.anim1.stop_property(widget, prop)
        self.anim2.stop_property(widget, prop)
        
        if not self.anim1.have_properties_to_animate(widget) and \
            not self.anim2.have_properties_to_animate(widget):
            self.stop(widget)

    def cancel(self, widget):
        self.anim1.cancel(widget)
        self.anim2.cancel(widget)
        super().cancel(widget)

    def cancel_property(self, widget, prop):
        '''Even if an animation is running, remove a property. It will not be
        animated further. If it was the only/last property being animated,
        the animation will be canceled (see :attr:`cancel`)
        This method overrides `:class:kivy.animation.Animation`'s
        version, to cancel it on all animations of the Sequence.
        .. versionadded:: 1.10.0
        '''
        self.anim1.cancel_property(widget, prop)
        self.anim2.cancel_property(widget, prop)
        if not self.anim1.have_properties_to_animate(widget) and \
            not self.anim2.have_properties_to_animate(widget):
            self.cancel(widget)

    def have_properties_to_animate(self, widget):
        return (self.anim1.have_properties_to_animate(widget) or
                self.anim2.have_properties_to_animate(widget))

    @property
    def animated_properties(self):
        return ChainMap({},
                        self.anim2.animated_properties,
                        self.anim1.animated_properties)

    @property
    def transition(self):
        # This property is impossible to implement
        raise AttributeError(
            "Can't lookup transition attribute of a CompoundAnimation")


class Sequence(CompoundAnimation):

    def on_anim1_complete(self, instance, widget):
        if widget.uid not in self._widgets:
            return None
        
        self.anim2.start(widget)

    def on_anim1_progress(self, instance, widget, progress):
        self.dispatch('on_progress', widget, progress / 2.)

    def on_anim2_complete(self, instance, widget):
        '''Repeating logic used with boolean variable "repeat".
        .. versionadded:: 1.7.1
        '''
        if widget.uid not in self._widgets:
            return None
        
        if self.repeat:
            self.anim1.start(widget)
        else:
            self.dispatch('on_complete', widget)
            self.cancel(widget)

    def on_anim2_progress(self, instance, widget, progress):
        self.dispatch('on_progress', widget, 0.5 + progress / 2.0)

    def __init__(self, anim1, anim2):
        super().__init__()

        #: Repeat the sequence. See 'Repeating animation' in the header
        #: documentation.
        self.repeat = False

        self.anim1 = anim1
        self.anim2 = anim2

        self.anim1.bind(on_complete=self.on_anim1_complete,
                        on_progress=self.on_anim1_progress)
        
        self.anim2.bind(on_complete=self.on_anim2_complete,
                        on_progress=self.on_anim2_progress)

    @property
    def duration(self):
        return self.anim1.duration + self.anim2.duration

    def stop(self, widget):
        props = self._widgets.pop(widget.uid, None)
        self.anim1.stop(widget)
        self.anim2.stop(widget)
        
        if props:
            self.dispatch('on_complete', widget)
        
        super().cancel(widget)

    def start(self, widget):
        self.stop(widget)
        self._widgets[widget.uid] = True
        self._register()
        self.dispatch('on_start', widget)
        self.anim1.start(widget)


class Parallel(CompoundAnimation):

    def on_anim_complete(self, instance, widget):
        self._widgets[widget.uid]['complete'] += 1
        
        if self._widgets[widget.uid]['complete'] == 2:
            self.stop(widget)
    
    def on_anim_progress(self, instance, widget, progress):
        self.dispatch('on_progress', widget, progress)

    def __init__(self, anim1, anim2):
        super().__init__()
        self.anim1 = anim1
        self.anim2 = anim2

        self.anim1.bind(on_complete=self.on_anim_complete)
        self.anim2.bind(on_complete=self.on_anim_complete)

        if self.anim1.duration > self.anim2.duration:
            self.anim1.bind(on_progress=self.on_anim_progress)
        else:
            self.anim2.bind(on_progress=self.on_anim_progress)

    @property
    def duration(self):
        return max(self.anim1.duration, self.anim2.duration)

    def stop(self, widget):
        self.anim1.stop(widget)
        self.anim2.stop(widget)
        
        if self._widgets.pop(widget.uid, None):
            self.dispatch('on_complete', widget)
        
        super().cancel(widget)

    def start(self, widget):
        self.stop(widget)
        self.anim1.start(widget)
        self.anim2.start(widget)
        self._widgets[widget.uid] = {'complete': 0}
        self._register()
        self.dispatch('on_start', widget)



class AnimationTransition():
    '''Collection of animation functions to be used with the Animation object.
    Easing Functions ported to Kivy from the Clutter Project
    https://developer.gnome.org/clutter/stable/ClutterAlpha.html
    The `progress` parameter in each animation function is in the range 0-1.
    '''

    @staticmethod
    def linear(progress):
        '''.. image:: images/anim_linear.png'''
        return progress

    @staticmethod
    def in_quad(progress):
        '''.. image:: images/anim_in_quad.png
        '''
        return (progress * progress)

    @staticmethod
    def out_quad(progress):
        '''.. image:: images/anim_out_quad.png
        '''
        return ( -1.0 * progress * (progress - 2.0) )

    @staticmethod
    def in_out_quad(progress):
        '''.. image:: images/anim_in_out_quad.png
        '''
        p = (progress * 2)
        if p < 1:
            return 0.5 * p * p
        p -= 1.0
        return ( -0.5 * (p * (p - 2.0) - 1.0) )

    @staticmethod
    def in_cubic(progress):
        '''.. image:: images/anim_in_cubic.png
        '''
        return (progress * progress * progress)

    @staticmethod
    def out_cubic(progress):
        '''.. image:: images/anim_out_cubic.png
        '''
        p = progress - 1.0
        return (p * p * p + 1.0)

    @staticmethod
    def in_out_cubic(progress):
        '''.. image:: images/anim_in_out_cubic.png
        '''
        p = progress * 2
        if p < 1:
            return (0.5 * p * p * p)
        p -= 2
        return ( 0.5 * (p * p * p + 2.0) )

    @staticmethod
    def in_quart(progress):
        '''.. image:: images/anim_in_quart.png
        '''
        return (progress * progress * progress * progress)

    @staticmethod
    def out_quart(progress):
        '''.. image:: images/anim_out_quart.png
        '''
        p = progress - 1.0
        return ( -1.0 * (p * p * p * p - 1.0) )

    @staticmethod
    def in_out_quart(progress):
        '''.. image:: images/anim_in_out_quart.png
        '''
        p = progress * 2
        if p < 1:
            return (0.5 * p * p * p * p)
        p -= 2
        return ( -0.5 * (p * p * p * p - 2.0) )

    @staticmethod
    def in_quint(progress):
        '''.. image:: images/anim_in_quint.png
        '''
        return (progress * progress * progress * progress * progress)

    @staticmethod
    def out_quint(progress):
        '''.. image:: images/anim_out_quint.png
        '''
        p = progress - 1.0
        return (p * p * p * p * p + 1.0)

    @staticmethod
    def in_out_quint(progress):
        '''.. image:: images/anim_in_out_quint.png
        '''
        p = progress * 2
        if p < 1:
            return (0.5 * p * p * p * p * p)
        p -= 2.0
        return ( 0.5 * (p * p * p * p * p + 2.0) )

    @staticmethod
    def in_sine(progress):
        '''.. image:: images/anim_in_sine.png
        '''
        return (-1.0 * cos(progress * (pi / 2.0)) + 1.0)

    @staticmethod
    def out_sine(progress):
        '''.. image:: images/anim_out_sine.png
        '''
        return sin(progress * (pi / 2.0))

    @staticmethod
    def in_out_sine(progress):
        '''.. image:: images/anim_in_out_sine.png
        '''
        return ( -0.5 * (cos(pi * progress) - 1.0) )

    @staticmethod
    def in_expo(progress):
        '''.. image:: images/anim_in_expo.png
        '''
        if progress == 0:
            return 0.0
        return pow(2, 10 * (progress - 1.0))

    @staticmethod
    def out_expo(progress):
        '''.. image:: images/anim_out_expo.png
        '''
        if progress == 1.0:
            return 1.0
        return -pow(2, -10 * progress) + 1.0

    @staticmethod
    def in_out_expo(progress):
        '''.. image:: images/anim_in_out_expo.png
        '''
        if progress == 0:
            return 0.0
        
        if progress == 1.0:
            return 1.0
        
        p = progress * 2
        if p < 1:
            return ( 0.5 * pow(2, 10 * (p - 1.0)) )
        
        p -= 1.0
        return ( 0.5 * (-pow(2, -10 * p) + 2.0) )

    @staticmethod
    def in_circ(progress):
        '''.. image:: images/anim_in_circ.png
        '''
        return ( -1.0 * (sqrt(1.0 - progress * progress) - 1.0) )

    @staticmethod
    def out_circ(progress):
        '''.. image:: images/anim_out_circ.png
        '''
        p = progress - 1.0
        return sqrt(1.0 - p * p)

    @staticmethod
    def in_out_circ(progress):
        '''.. image:: images/anim_in_out_circ.png
        '''
        p = progress * 2
        if p < 1:
            return ( -0.5 * (sqrt(1.0 - p * p) - 1.0) )
        
        p -= 2.0
        return ( 0.5 * (sqrt(1.0 - p * p) + 1.0) )

    @staticmethod
    def in_elastic(progress):
        '''.. image:: images/anim_in_elastic.png
        '''
        p = .3
        s = p / 4.0
        q = progress
        if q == 1:
            return 1.0
        q -= 1.0
        return -(pow(2, 10 * q) * sin((q - s) * (2 * pi) / p))

    @staticmethod
    def out_elastic(progress):
        '''.. image:: images/anim_out_elastic.png
        '''
        p = 0.3
        s = p / 4.0
        q = progress
        if q == 1:
            return 1.0
        return (pow(2, -10 * q) * sin((q - s) * (2 * pi) / p) + 1.0)

    @staticmethod
    def in_out_elastic(progress):
        '''.. image:: images/anim_in_out_elastic.png
        '''
        p = .3 * 1.5
        s = p / 4.0
        q = progress * 2
        if q == 2:
            return 1.0
        if q < 1:
            q -= 1.0
            return ( -0.5 * (pow(2, 10 * q) * sin((q - s) * (2.0 * pi) / p)) )
        else:
            q -= 1.0
            return (pow(2, -10 * q) * sin((q - s) * (2.0 * pi) / p) * 0.5 + 1.0)

    @staticmethod
    def in_back(progress):
        '''.. image:: images/anim_in_back.png
        '''
        return ( progress * progress * ((1.70158 + 1.0) * progress - 1.70158) )

    @staticmethod
    def out_back(progress):
        '''.. image:: images/anim_out_back.png
        '''
        p = progress - 1.0
        return (p * p * ((1.70158 + 1) * p + 1.70158) + 1.0)

    @staticmethod
    def in_out_back(progress):
        '''.. image:: images/anim_in_out_back.png
        '''
        p = progress * 2.
        s = 1.70158 * 1.525
        if p < 1:
            return ( 0.5 * (p * p * ((s + 1.0) * p - s)) )
        
        p -= 2.0
        return ( 0.5 * (p * p * ((s + 1.0) * p + s) + 2.0) )

    @staticmethod
    def _out_bounce_internal(t, d):
        p = t / d

        if p < (1.0 / 2.75):
            return (7.5625 * p * p)
        
        elif p < (2.0 / 2.75):
            p -= (1.5 / 2.75)
            return (7.5625 * p * p + 0.75)
        
        elif p < (2.5 / 2.75):
            p -= (2.25 / 2.75)
            return (7.5625 * p * p + 0.9375)
        
        else:
            p -= (2.625 / 2.75)
            return (7.5625 * p * p + 0.984375)

    @staticmethod
    def _in_bounce_internal(t, d):
        return ( 1.0 - AnimationTransition._out_bounce_internal(d - t, d) )

    @staticmethod
    def in_bounce(progress):
        '''.. image:: images/anim_in_bounce.png
        '''
        return AnimationTransition._in_bounce_internal(progress, 1.0)

    @staticmethod
    def out_bounce(progress):
        '''.. image:: images/anim_out_bounce.png
        '''
        return AnimationTransition._out_bounce_internal(progress, 1.0)

    @staticmethod
    def in_out_bounce(progress):
        '''.. image:: images/anim_in_out_bounce.png
        '''
        p = progress * 2.0
        if p < 1.0:
            return ( AnimationTransition._in_bounce_internal(p, 1.0) * 0.5 )
        
        return ( AnimationTransition._out_bounce_internal(p - 1.0, 1.0) * 0.5 + 0.5 )




from kivy.animation import Animation
from kivy.input.providers.mouse import MouseMotionEvent
from kivy.clock import Clock
from kivy.lang import Builder
from kivy.metrics import dp, sp
from kivy.properties import (
	ListProperty, NumericProperty,
	ObjectProperty, StringProperty,
	BooleanProperty
)
from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.textinput import TextInput


Builder.load_string("""

#:import ToggleButtonIcon kivygo.uix.icon.ToggleButtonIcon
#:import InputEditor kivygo.uix.terminal.InputEditor


<IconInput>:
	hide: False
	padding: [dp(10), dp(10), dp(10), dp(10)]
	size_hint_y: None
	height: '60dp'
	input: input

	BoxLayout:
		id: box
		canvas:
			Color:
				rgba: root.color_line
			Line:
				rounded_rectangle: (self.pos + self.size + root.radius + [100])
				width: root.outline_width
		canvas.before:
			Color:
				rgba: root.background_color
			RoundedRectangle:
				size: self.size
				pos: self.pos
				radius: root.radius

		AnchorLayout:
			id: anchor_left
			padding: [1, 1, 1, 1]
			size_hint_x: None
			width: ( root.icon_left_size[0] + dp(15) )
			anchor_y: 'center'
			
			ToggleButtonIcon:
				id: button_left
				allow_stretch: True
				name: 'icon_left'
				keep_ratio: False
				mipmap: True
				size_hint_y: None
				size: root.icon_left_size
				pos_source: root.icon_left_pos_source
				icon_source: root.icon_left_source
				icon_color: root.icon_left_color
				pos_color: root.icon_left_color_pos
				icon_state_source: root.icon_left_state_sources
				effect_color: root.icon_left_effect_color
				state_button: root.icon_left_state

		AnchorLayout:
			id: anchor_input

			AnchorLayout:
				anchor_y: 'center'
				anchor_x: 'center'
				padding: root.padding
				MyTextInput:
				# InputEditor:
					id: input
					size_hint_y: None
					height: self.minimum_height			
					id: input
					text: root.input_text
					on_text: root.input_text = self.text
					window_root: root
					background_color: [1, 1, 1, 0]
					password: root.hide
					foreground_color: root.text_input_color
					multiline: root.multiline
					# height: min(box.height, self.input_minimum_height)
					# change_width: False
					# bar_color: [0, 0, 0, 0]
					# bar_inactive_color: [0, 0, 0, 0]
					# on_text_validate: root.dispatch('on_enter')

		AnchorLayout:
			padding: [dp(-1), dp(1), dp(10), dp(1)]
			size_hint_x: None
			width: ( root.icon_right_size[0] + dp(15) )
			anchor_y: 'center'
			id: anchor_right

			ToggleButtonIcon:
				id: button_right
				name: 'icon_right'
				window_root: root
				allow_stretch: True
				keep_ratio: False
				mipmap: True
				size_hint_y: None
				size: root.icon_right_size
				pos_source: root.icon_right_pos_source
				image_source: root.icon_right_source
				icon_state_source: root.icon_right_state_sources
				icon_color: root.icon_right_color
				pos_color: root.icon_right_color_pos
				effect_color: root.icon_right_effect_color
				state_button: root.icon_right_state
	
	FloatLayout:
		id: float_lbl
		size_hint_x: None
		x: root.x
		width: lbl.texture_size[0] + dp('140')

		Label:
			id: lbl
			text: root.label_text
			font_size: root.label_font_size
			color: root.label_defaut_color

""")


class MyTextInput(TextInput):
	window_root = ObjectProperty(None)
	def insert_text(self, substring, from_undo=False):
		r =  super().insert_text(substring, from_undo=from_undo)

		if self.window_root != None:
			self.window_root.dispatch('on_input_text', substring, from_undo)
		return r


class IconInput(AnchorLayout):

	line_color = ListProperty([1,1,1,1])
	line_color_pos = ListProperty([0, 0, 0, 0])
	color_line = ListProperty([0, 0, 0, 0])
	outline_width = NumericProperty(1.01)

	background_color = ListProperty([0,0,0,0])
	multiline = BooleanProperty(False)
	radius = ListProperty([dp(15), dp(15), dp(15), dp(15)])

	icon_left = ObjectProperty(None)
	icon_left_state = StringProperty('toggle')#'toggle' or 'button'
	icon_left_color = ListProperty([1, 1, 1, 1])
	icon_left_pos_source = StringProperty('')
	icon_left_source = StringProperty('')
	icon_left_color_pos = ListProperty([-1, -1, -1, -1])
	icon_left_state_sources = ListProperty(['', ''])
	icon_left_size = ListProperty([dp(30), dp(23)])
	icon_left_effect_color = ListProperty([0, 0, 0, 0])

	icon_right = ObjectProperty(None)
	icon_right_state = StringProperty('toggle')#'toggle' or 'button'
	icon_right_color = ListProperty([1, 1, 1, 1])
	icon_right_pos_source = StringProperty('')
	icon_right_state_sources = ListProperty(['', ''])
	icon_right_source = StringProperty('')
	icon_right_color_pos = ListProperty([-1, -1, -1, -1])
	icon_right_size = ListProperty([dp(30), dp(23)])
	icon_right_effect_color = ListProperty([0, 0, 0, 0])

	label_text = StringProperty('')
	label_font_size = NumericProperty(dp(16))
	label_defaut_color = ListProperty([1, 1, 1, 1])
	label_pos_color = ListProperty([1, 1, 1, 1])
	state_label = StringProperty('')

	input = ObjectProperty(None)
	hide = ObjectProperty(False)
	text_input_color = ListProperty([1, 1, 1, 1])
	input_text = StringProperty("")

	__events__ = ('on_icon_right_press', 'on_icon_right_release', 'on_icon_left_press',
				  'on_icon_left_release', 'on_input_press', 'on_input_release', 
				  'on_input_text', 'on_init_input', 'on_icon_right_pos',
				  'on_icon_right_pos_release', 'on_enter')

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.pdentro = (0, 0)
		self.pfora = (0, 0)
		Clock.schedule_once(self.config)

	def on_icon_right_state(self, *args):
		pass
	
	def on_icon_left_state(self, *args):
		pass


	def select(self, *args):
		self.input.focus = True
		touch = MouseMotionEvent(None, 123, args=(1, 1))  # args are device, id, spos
		touch.button = 'left'
		touch.pos = [self.center_x, self.center_y]
		self.dispatch('on_touch_down', touch)
	
	def un_select(self, *args):
		self.input.focus = False
		touch = MouseMotionEvent(None, 123, args=(1, 1))  # args are device, id, spos
		touch.button = 'left'
		touch.pos = [self.center_x, self.center_y]
		self.dispatch('on_touch_up', touch)


	def icon_down(self, button, *args):
		if button.state_button == 'toggle':
			button.num += 1
			if button.pos_source and button.num == 1:
				button.source = button.pos_source
			elif button.num == 2:
				button.source = button.icon_source
				button.num = 0
		elif button.state_button == 'button':
			if button.pos_source:
				button.source = button.pos_source
	
	def icon_up(self, button, *args):
		if button.state_button == 'button':
			button.source = button.icon_source

	def config(self, *args):
		self.pads()
		self.color_line = self.line_color
		self.icon_right = self.ids.button_right
		self.icon_left = self.ids.button_left
		self.input = self.ids.input

		if not self.icon_left_source and not self.icon_left_state_sources[0] and not self.icon_left_pos_source:
			self.ids.box.remove_widget(self.ids.anchor_left)
			self.icon_left_size = [0, 0]
			self.icon_left = False
		if not self.icon_right_source and not self.icon_right_state_sources[0] and not self.icon_right_pos_source:
			self.ids.box.remove_widget(self.ids.anchor_right)
			self.icon_right_size = [0,0]
			self.icon_right = False
		self.dispatch('on_init_input')

	def pads(self, *args):
		radius_left = self.radius[0] if self.radius[0] > self.radius[3] else self.radius[3] 
		radius_right = self.radius[1] if self.radius[1] > self.radius[2] else self.radius[2]
		
		if radius_left > 13:
			one_pad_x = radius_left-radius_left/2
			two_pad_x = -one_pad_x/1.4
			self.ids.anchor_left.padding = [one_pad_x, 1, two_pad_x, 1]
			self.ids.anchor_input.padding = [one_pad_x, 1, one_pad_x, 1]
		else:
			self.ids.anchor_left.padding = [7, 1, 15, 1]
		
		if radius_right > dp(13):
			one_pad_x = radius_right-radius_right/2
			two_pad_x = -one_pad_x/1.7
			self.ids.anchor_right.padding = [two_pad_x, 1, one_pad_x, 1]
			# self.ids.anchor_input.padding = [one_pad_x, 1, two_pad_x, 1]
		else:
			self.ids.anchor_left.padding = [7, 1, -3, 1]
			# self.ids.anchor_input.padding = [5, 1, 2, 1]

	def on_touch_down(self, touch):
		if touch.is_mouse_scrolling:
			return False

		if self.ids.anchor_input.collide_point(*touch.pos):
			self.anima(True)
			self.dispatch('on_input_press')
			
		if self.icon_right:
			if self.icon_right.collide_point(*touch.pos):
				# self.icon_down(self.icon_right)
				self.dispatch('on_icon_right_press')
		if self.icon_left:
			if self.icon_left.collide_point(*touch.pos):
				#se .icon_down(self.icon_left)
				self.dispatch('on_icon_left_press')

		return super().on_touch_down(touch)
	
	def on_touch_up(self, touch):
		if touch.is_mouse_scrolling:
			return False

		if not self.ids.anchor_input.collide_point(*touch.pos):
			self.dispatch('on_input_release')
			if self.ids.input.text == '':
				self.anima(False)
		if self.icon_right:
			if self.icon_right.collide_point(*touch.pos):
				# self.icon_up(self.icon_right)
				self.dispatch('on_icon_right_release')
		if self.icon_left:
			if self.icon_left.collide_point(*touch.pos):
				# self.icon_up(self.icon_left)
				self.dispatch('on_icon_left_release')
		return super().on_touch_up(touch)

	def on_icon_right_press(self): pass
	def on_icon_right_release(self): pass
	def on_icon_right_pos(self, *args): pass
	def on_icon_right_pos_release(self, *args): pass
	
	def on_icon_left_press(self): pass
	def on_icon_left_release(self): pass

	def on_input_press(self): pass
	def on_input_release(self): pass

	def on_init_input(self): pass
	def on_input_text(self, substring, from_undo): pass
	def on_enter(self, *args): pass

	def on_pos(self, *args):
		self.pdentro = (self.x-dp(50)+self.radius[-1]/2.5, self.y+dp(40))
		radius_left = self.radius[0] if self.radius[0] > self.radius[3] else self.radius[3] 
		if radius_left <= dp(13):
			radius_left = dp(13)
		if self.icon_left_size[0] == 0:
			self.pfora = (self.x-dp(40)+radius_left, self.y+dp(10))
		else:
			self.pfora = (self.x-dp(40)+self.icon_left_size[0]+(radius_left/2), self.y+dp(10))

		if self.state_label == 'pdentro':
			self.ids.lbl.pos = self.pdentro
		elif self.state_label == 'pfora':
			self.ids.lbl.pos = self.pfora
		else:
			self.ids.lbl.pos = self.pfora

	def anima(self, pos_widget, *args):
		if pos_widget:
			self.state_label = 'pdentro'
			newpos = self.pdentro
			font = sp(12)
			if self.line_color_pos != [0, 0, 0, 0]:
				self.color_line = self.line_color_pos
			self.ids.lbl.color = self.label_pos_color
		if not pos_widget:
			self.state_label = 'pfora'
			newpos = self.pfora
			font = sp(16)
			self.color_line = self.line_color
			self.ids.lbl.color = self.label_defaut_color
		
		anim = Animation(pos=(newpos), font_size=font, d=.1, t='out_sine')
		anim.start(self.ids.lbl)





from kivy.animation import Animation
from kivy.clock import Clock
from kivygo.uix.circularlayout import CircularLayout
from kivy.graphics import Line, Color, Ellipse
from kivy.lang import Builder
from kivy.properties import (
    NumericProperty, BoundedNumericProperty,
    ObjectProperty, StringProperty, AliasProperty,
    ListProperty, ReferenceListProperty,
    OptionProperty, BooleanProperty,
    DictProperty
)
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label

from math import atan, pi, radians, sin, cos
import datetime


def map_number(x, in_min, in_max, out_min, out_max):
    return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min

def rgb_to_hex(*color):
    tor = "#"
    for c in color:
        tor += "{:>02}".format(hex(int(c*255))[2:])
    return tor


Builder.load_string("""

<Number>:
    text_size: self.size
    valign: "middle"
    halign: "center"
    font_size: self.height * self.size_factor

<CircularNumberPicker>:
    canvas.before:
        PushMatrix
        Scale:
            origin: self.center_x + self.padding[0] - self.padding[2], self.center_y + self.padding[3] - self.padding[1]
            x: self.scale
            y: self.scale

    canvas.after:
        PopMatrix            

<CircularTimePicker>:
    orientation: "vertical"
    spacing: "20dp"

    AnchorLayout:
        anchor_x: "center"
        anchor_y: "center"
        size_hint_y: 1.0 / 3

        GridLayout:
            cols: 2
            spacing: "10dp"
            size_hint_x: None
            width: self.minimum_width
           
            Label:
                id: timelabel
                text: root.time_text
                markup: True
                halign: "right"
                valign: "middle"
                size_hint_x: None
                width: self.texture_size[0]
                font_size: self.height * 0.75

            Label:
                id: ampmlabel
                text: root.ampm_text
                markup: True
                halign: "left"
                valign: "middle"
                size_hint_x: None
                width: self.texture_size[0]
                font_size: self.height * 0.3

    FloatLayout:
        id: picker_container
        _bound: {}
""")


class Number(Label):
    """The class used to show the numbers in the selector.
    """

    size_factor = NumericProperty(.5)
    """Font size scale.
    """


class CircularNumberPicker(CircularLayout):
    """A circular number picker based on CircularLayout. A selector will
    help you pick a number. You can also set :attr:`multiples_of` to make
    it show only some numbers and use the space in between for the other
    numbers.
    """

    min = NumericProperty(0)
    """The first value of the range.
    """

    max = NumericProperty(0)
    """The last value of the range. Note that it behaves like range, so
    the actual last displayed value will be :attr:`max` - 1.
    """

    range = ReferenceListProperty(min, max)
    """Packs :attr:`min` and :attr:`max` into a list for convenience. See
    their documentation for further information.
    """

    multiples_of = NumericProperty(1)
    """Only show numbers that are multiples of this number. The other numbers
    will be selectable, but won't have their own label.
    """

    selector_color = ListProperty([0.337, 0.439, 0.490])
    """Color of the number selector. RGB.
    """

    color = ListProperty([1, 1, 1])
    """Color of the number labels and of the center dot. RGB.
    """

    selector_alpha = BoundedNumericProperty(0.3, min=0, max=1)
    """Alpha value for the transparent parts of the selector.
    """

    selected = NumericProperty(None)
    """Currently selected number. defaults to :attr:`min`.
    """

    number_size_factor = NumericProperty(0.5)
    """Font size scale factor fot the :class:`Number`s.
    """

    number_format_string = StringProperty("{}")
    """String that will be formatted with the selected number as the first argument.
    Can be anything supported by :meth:`str.format` (es. "{:02d}").
    """

    scale = NumericProperty(1)
    """Canvas scale factor. Used in :class:`CircularTimePicker` transitions.
    """

    _selection_circle = ObjectProperty(None)
    _selection_line = ObjectProperty(None)
    _selection_dot = ObjectProperty(None)
    _selection_dot_color = ObjectProperty(None)
    _selection_color = ObjectProperty(None)
    _center_dot = ObjectProperty(None)
    _center_color = ObjectProperty(None)


    def _get_items(self):
        return self.max - self.min
    
    items = AliasProperty(_get_items, None)


    def _get_shown_items(self):
        c = 0
        for i in range(*self.range):
            if i % self.multiples_of == 0:
                c += 1
        return c
    shown_items = AliasProperty(_get_shown_items, None)

    def __init__(self, **kw):
        self._trigger_genitems = Clock.create_trigger(self._genitems, -1)
        self.bind(
            min=self._trigger_genitems,
            max=self._trigger_genitems,
            multiples_of=self._trigger_genitems
        )
        
        super().__init__(**kw)
        
        self.selected = self.min
        self.bind(
            selected=self.on_selected,
            pos=self.on_selected,
            size=self.on_selected
        )

        cx = self.center_x + self.padding[0] - self.padding[2]
        cy = self.center_y + self.padding[3] - self.padding[1]
        sx, sy = self.pos_for_number(self.selected)
        epos = [i - (self.delta_radii * self.number_size_factor) for i in (sx, sy)]
        esize = [self.delta_radii * self.number_size_factor * 2]*2
        dsize = [i * .3 for i in esize]
        dpos = [i + esize[0] / 2. - dsize[0] / 2. for i in epos]
        csize = [i * .05 for i in esize]
        cpos = [i - csize[0] / 2. for i in (cx, cy)]
        dot_alpha = 0 if self.selected % self.multiples_of == 0 else 1
        color = list(self.selector_color)

        with self.canvas:
            self._selection_color = Color(*(color + [self.selector_alpha]))
            self._selection_circle = Ellipse(pos=epos, size=esize)
            self._selection_line = Line(points=[cx, cy, sx, sy])
            self._selection_dot_color = Color(*(color + [dot_alpha]))
            self._selection_dot = Ellipse(pos=dpos, size=dsize)
            self._center_color = Color(*self.color)
            self._center_dot = Ellipse(pos=cpos, size=csize)

        self.bind(selector_color=lambda ign, c: setattr(self._selection_color, "rgba", c + [self.selector_alpha]))
        self.bind(selector_color=lambda ign, c: setattr(self._selection_dot_color, "rgb", c))
        self.bind(color=lambda ign, c: setattr(self._center_color, "rgb", c))
        Clock.schedule_once(self._genitems)
        Clock.schedule_once(self.on_selected) # Just to make sure pos/size are set

    def _genitems(self, *a):
        self.clear_widgets()
        for i in range(*self.range):
            if i % self.multiples_of != 0:
                continue
            n = Number(text=self.number_format_string.format(i), size_factor=self.number_size_factor, color=self.color)
            self.bind(color=n.setter("color"))
            self.add_widget(n)

    def on_touch_down(self, touch):
        if not self.collide_point(*touch.pos):
            return None
        
        touch.grab(self)
        self.selected = self.number_at_pos(*touch.pos)

    def on_touch_move(self, touch):
        if touch.grab_current is not self:
            return super().on_touch_move(touch)
        self.selected = self.number_at_pos(*touch.pos)

    def on_touch_up(self, touch):
        if touch.grab_current is not self:
            return super().on_touch_up(touch)
        touch.ungrab(self)

    def on_selected(self, *a):
        cx = self.center_x + self.padding[0] - self.padding[2]
        cy = self.center_y + self.padding[3] - self.padding[1]
        sx, sy = self.pos_for_number(self.selected)
        epos = [i - (self.delta_radii * self.number_size_factor) for i in (sx, sy)]
        esize = [self.delta_radii * self.number_size_factor * 2]*2
        dsize = [i * .3 for i in esize]
        dpos = [i + esize[0] / 2. - dsize[0] / 2. for i in epos]
        csize = [i * .05 for i in esize]
        cpos = [i - csize[0] / 2. for i in (cx, cy)]
        dot_alpha = 0 if self.selected % self.multiples_of == 0 else 1

        if self._selection_circle:
            self._selection_circle.pos = epos
            self._selection_circle.size = esize
        
        if self._selection_line:
            self._selection_line.points = [cx, cy, sx, sy]
        
        if self._selection_dot:
            self._selection_dot.pos = dpos
            self._selection_dot.size = dsize
        
        if self._selection_dot_color:
            self._selection_dot_color.a = dot_alpha
        
        if self._center_dot:
            self._center_dot.pos = cpos
            self._center_dot.size = csize

    def pos_for_number(self, n):
        """Returns the center x, y coordinates for a given number.
        """
        if self.items == 0:
            return [0, 0]
        
        radius = min(self.width-self.padding[0]-self.padding[2], self.height-self.padding[1]-self.padding[3]) / 2.
        middle_r = radius * sum(self.radius_hint) / 2.
        cx = self.center_x + self.padding[0] - self.padding[2]
        cy = self.center_y + self.padding[3] - self.padding[1]
        sign = +1.
        angle_offset = radians(self.start_angle)
        if self.direction == 'cw':
            angle_offset = 2 * pi - angle_offset
            sign = -1.
        quota = 2*pi / self.items
        mult_quota = 2*pi / self.shown_items
        angle = angle_offset + n * sign * quota

        if self.items == self.shown_items:
            angle += quota / 2
        else:
            angle -= mult_quota / 2

        # kived: looking it up, yes. x = cos(angle) * radius + centerx; y = sin(angle) * radius + centery
        x = cos(angle) * middle_r + cx
        y = sin(angle) * middle_r + cy

        return [x, y]

    def number_at_pos(self, x, y):
        """Returns the number at a given x, y position. The number is found
        using the widget's center as a starting point for angle calculations.

        Not thoroughly tested, may yield wrong results.
        """
        if self.items == 0:
            return self.min

        cx = self.center_x + self.padding[0] - self.padding[2]
        cy = self.center_y + self.padding[3] - self.padding[1]
        lx = x - cx
        ly = y - cy
        quota = 2*pi / self.items
        mult_quota = 2*pi / self.shown_items
        if lx == 0 and ly > 0:
            angle = pi/2
        elif lx == 0 and ly < 0:
            angle = 3*pi/2
        else:
            angle = atan(ly/lx)
            if lx < 0 and ly > 0:
                angle += pi
            if lx > 0 and ly < 0:
                angle += 2*pi
            if lx < 0 and ly < 0:
                angle += pi
        angle += radians(self.start_angle)
        if self.direction == "cw":
            angle = 2*pi - angle
        if mult_quota != quota:
            angle -= mult_quota / 2
        if angle < 0:
            angle += 2*pi
        elif angle > 2*pi:
            angle -= 2*pi

        return min(int(angle / quota) + self.min, self.max-1)


class CircularMinutePicker(CircularNumberPicker):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.min = 0
        self.max = 60
        self.multiples_of = 5
        self.number_format_string = "{:02d}"
        self.direction = "cw"
        self.bind(shown_items=self._update_start_angle)
        Clock.schedule_once(self._update_start_angle)
        Clock.schedule_once(self.on_selected)

    def _update_start_angle(self, *a):
        self.start_angle = -(360. / self.shown_items / 2) - 90


class CircularHourPicker(CircularNumberPicker):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.min = 1
        self.max = 13
        self.multiples_of = 1
        self.number_format_string = "{}"
        self.direction = "cw"
        self.bind(shown_items=self._update_start_angle)
        Clock.schedule_once(self._update_start_angle)
        Clock.schedule_once(self.on_selected)

    def _update_start_angle(self, *a):
        self.start_angle = (360.0 / self.shown_items / 2) - 90


class CircularTimePicker(BoxLayout):
    """Widget that makes use of :class:`CircularHourPicker` and
    :class:`CircularMinutePicker` to create a user-friendly, animated
    time picker like the one seen on Android.

    See module documentation for more details.
    """

    hours = NumericProperty(0)
    """The hours, in military format (0-23).

    :attr:`hours` is a :class:`~kivy.properties.NumericProperty` and
    defaults to 0 (12am).
    """

    minutes = NumericProperty(0)
    """The minutes.
    """

    time_list = ReferenceListProperty(hours, minutes)
    """Packs :attr:`hours` and :attr:`minutes` in a list for convenience.
    """

    time_format = StringProperty("[color={hours_color}][ref=hours]{hours}[/ref][/color]:[color={minutes_color}][ref=minutes]{minutes:02d}[/ref][/color]")
    """String that will be formatted with the time and shown in the time label.
    Can be anything supported by :meth:`str.format`. Make sure you don't
    remove the refs. See the default for the arguments passed to format.

    :attr:`time_format` is a :class:`~kivy.properties.StringProperty` and
    defaults to "[color={hours_color}][ref=hours]{hours}[/ref][/color]:[color={minutes_color}][ref=minutes]{minutes:02d}[/ref][/color]".
    """

    ampm_format = StringProperty("[color={am_color}][ref=am]AM[/ref][/color]\n[color={pm_color}][ref=pm]PM[/ref][/color]")
    """String that will be formatted and shown in the AM/PM label.
    Can be anything supported by :meth:`str.format`. Make sure you don't
    remove the refs. See the default for the arguments passed to format.

    :attr:`ampm_format` is a :class:`~kivy.properties.StringProperty` and
    defaults to "[color={am_color}][ref=am]AM[/ref][/color]\n[color={pm_color}][ref=pm]PM[/ref][/color]".
    """

    picker = OptionProperty("hours", options=("minutes", "hours"))
    """Currently shown time picker. Can be one of "minutes", "hours".
    """

    selector_color = ListProperty([0.337, 0.439, 0.490])
    """Color of the number selector and of the highlighted text. RGB.
    """

    color = ListProperty([1, 1, 1])
    """Color of the number labels and of the center dot. RGB.
    """

    selector_alpha = BoundedNumericProperty(0.3, min=0, max=1)
    """Alpha value for the transparent parts of the selector.
    """

    _am = BooleanProperty(True)
    _h_picker = ObjectProperty(None)
    _m_picker = ObjectProperty(None)
    _bound = DictProperty({})


    def _get_time(self):
        return datetime.time(*self.time_list)
    
    def _set_time(self, dt):
        self.time_list = [dt.hour, dt.minute]
    
    time = AliasProperty(_get_time, _set_time, bind=("time_list",))
    """Selected time as a datetime.time object.
    """


    def _get_picker(self):
        if self.picker == "hours":
            return self._h_picker
        return self._m_picker
    
    _picker = AliasProperty(_get_picker, None)


    def _get_time_text(self):
        hc = rgb_to_hex(*self.selector_color) if self.picker == "hours" else rgb_to_hex(*self.color)
        mc = rgb_to_hex(*self.selector_color) if self.picker == "minutes" else rgb_to_hex(*self.color)
        h = self.hours == 0 and 12 or self.hours <= 12 and self.hours or self.hours - 12
        m = self.minutes
        return self.time_format.format(hours_color=hc, minutes_color=mc, hours=h, minutes=m)
    
    time_text = AliasProperty(_get_time_text, None, bind=("hours", "minutes", "time_format", "picker"))


    def _get_ampm_text(self):
        amc = rgb_to_hex(*self.selector_color) if self._am else rgb_to_hex(*self.color)
        pmc = rgb_to_hex(*self.selector_color) if not self._am else rgb_to_hex(*self.color)
        return self.ampm_format.format(am_color=amc, pm_color=pmc)
    
    ampm_text = AliasProperty(_get_ampm_text, None, bind=("hours", "ampm_format", "_am"))


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.hours >= 12:
            self._am = False
        
        self.bind(time_list=self.on_time_list, picker=self._switch_picker, _am=self.on_ampm)
        self._h_picker = CircularHourPicker()
        self._m_picker = CircularMinutePicker()
        Clock.schedule_once(self.on_selected)
        Clock.schedule_once(self.on_time_list)
        Clock.schedule_once(self._init_later)
        Clock.schedule_once(lambda *a: self._switch_picker(noanim=True))
        
    def _init_later(self, *args):
        self.ids.timelabel.bind(on_ref_press=self.on_ref_press)
        self.ids.ampmlabel.bind(on_ref_press=self.on_ref_press)

    def on_ref_press(self, ign, ref):
        if ref == "hours":
            self.picker = "hours"
        
        elif ref == "minutes":
            self.picker = "minutes"
        
        elif ref == "am":
            self._am = True
        
        elif ref == "pm":
            self._am = False

    def on_selected(self, *a):
        if not self._picker:
            return None
        
        if self.picker == "hours":
            hours = self._picker.selected if self._am else self._picker.selected + 12
            if hours == 24 and not self._am:
                hours = 12
            
            elif hours == 12 and self._am:
                hours = 0
            
            self.hours = hours
        elif self.picker == "minutes":
            self.minutes = self._picker.selected

    def on_time_list(self, *a):
        if not self._picker:
            return None
        
        if self.picker == "hours":
            self._picker.selected = self.hours == 0 and 12 or self._am and self.hours or self.hours - 12
        
        elif self.picker == "minutes":
            self._picker.selected = self.minutes

    def on_ampm(self, *a):
        if self._am:
            self.hours = self.hours if self.hours <  12 else self.hours - 12
        else:
            self.hours = self.hours if self.hours >= 12 else self.hours + 12

    def _switch_picker(self, *args, **kwargs):
        noanim = "noanim" in kwargs
        if noanim:
            noanim = kwargs["noanim"]

        try:
            container = self.ids.picker_container
        except (AttributeError, NameError):
            Clock.schedule_once(lambda *a: self._switch_picker(noanim=noanim))

        if self.picker == "hours":
            picker = self._h_picker
            prevpicker = self._m_picker
        
        elif self.picker == "minutes":
            picker = self._m_picker
            prevpicker = self._h_picker

        if len(self._bound) > 0:
            prevpicker.unbind(selected=self.on_selected)
            self.unbind(**self._bound)
        
        picker.bind(selected=self.on_selected)
        self._bound = {
            "selector_color": picker.setter("selector_color"),
            "color":          picker.setter("color"),
            "selector_alpha": picker.setter("selector_alpha")
        }
        self.bind(**self._bound)
        
        if len(container._bound) > 0:
            container.unbind(**container._bound)
        
        container._bound = {"size": picker.setter("size"), "pos":  picker.setter("pos")}
        container.bind(**container._bound)

        picker.pos = container.pos
        picker.size = container.size
        picker.selector_color = self.selector_color
        picker.color = self.color
        picker.selector_alpha = self.selector_alpha

        if noanim:
            if prevpicker in container.children:
                container.remove_widget(prevpicker)
            
            if picker.parent:
                picker.parent.remove_widget(picker)
            
            container.add_widget(picker)
        else:
            if prevpicker in container.children:
                anim = Animation(scale=1.5, d=.5, t="in_back") & Animation(opacity=0, d=.5, t="in_cubic")
                anim.start(prevpicker)
                Clock.schedule_once(lambda *a: container.remove_widget(prevpicker), .5)
            
            picker.scale = 1.5
            picker.opacity = 0
            if picker.parent:
                picker.parent.remove_widget(picker)
            
            container.add_widget(picker)
            anim = Animation(scale=1, d=.5, t="out_back") & Animation(opacity=1, d=.5, t="out_cubic")
            Clock.schedule_once(lambda *a: anim.start(picker), .3)




from kivy.properties import ObjectProperty, ListProperty, StringProperty
from kivy.lang.builder import Builder
from kivy.clock import Clock
from kivy.metrics import dp

from kivy.uix.widget import Widget
from kivy.uix.image import Image

from kivy.core.image import Image as CoreImage
import tempfile

from svglib import svglib
from reportlab.graphics import renderPM
from reportlab.lib.colors import red, green
from PIL import ImageChops
from PIL import Image as PILImage
import io


def get_kivy_image_from_bytes(image_bytes, file_extension, image_widget=None):
	# Return a Kivy image set from a bytes variable
	buf = io.BytesIO(image_bytes)
	kw = dict()

	if image_widget != None:
		kw = {
			"mipmap": image_widget.mipmap,
			"anim_delay": image_widget.anim_delay,
			"keep_data": image_widget.keep_data,
			"nocache": image_widget.nocache
		}
	cim = CoreImage(buf, ext=file_extension, **kw)
	return cim


def drawing2png(path, fp):
	draw = svglib.svg2rlg(path)
	if draw == None:
		return None
	
	red_img = renderPM.drawToPIL(draw, bg=green, configPIL={'transparent': green})
	green_img = renderPM.drawToPIL(draw, bg=red, configPIL={'transparent': red})
	bg_mask = ImageChops.difference(red_img, green_img).convert('1', dither=PILImage.NONE)
	bg_mask = ImageChops.invert(bg_mask)
	red_img.putalpha(bg_mask)
	red_img.save(fp, fmt='png')


Builder.load_string("""

<RoudedImage>:
	size_hint_x: None
	width: '70dp'
	canvas:
		Color:
			rgba:[1, 1, 1, 1]
		RoundedRectangle:
			pos: self.pos
			size: self.size
			radius: root.radius
			texture: root.texture


""")


class RoudedImage(Widget):

	texture = ObjectProperty(None)
	source = StringProperty('')
	radius = ListProperty([dp(5), 0, 0, dp(5)])
	
	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		Clock.schedule_once(self.create_texture, 1)

	def create_texture(self, *args):
		image = Image(
			source=self.source, allow_stretch=True, keep_ratio=False, 
			size_hint=(None, None), size=self.size,
		)
		self.texture = image.texture


class ImageWithSVG(Image):
	
	image_source = ObjectProperty('')
	
	def on_image_source(self, *args):

		if isinstance(self.image_source, str):
			self.set_source(self.image_source)

	def set_source(self, source:str):

		if not source.endswith(".svg"):
			self.source = source
			return None
		
		file = tempfile.NamedTemporaryFile(mode='+wb', suffix=".png")
		try:
			drawing2png(source, file)
		except ValueError:
			pass

		file.seek(0)
		img = get_kivy_image_from_bytes(file.read(), 'png', self)
		self.texture = img.texture

